#
# Bitcoin Safe
# Copyright (C) 2024 Andreas Griffin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/gpl-3.0.html
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import logging
from typing import Dict, List, Optional, Type

import bdkpython as bdk
from bitcoin_qr_tools.data import Data

from bitcoin_safe.gui.qt.export_data import (
    DataGroupBox,
    ExportDataSimple,
    HorizontalImportExportGroups,
)
from bitcoin_safe.gui.qt.keystore_ui import SignedUI, SignerUI
from bitcoin_safe.gui.qt.step_progress_bar import StepProgressContainer
from bitcoin_safe.signals import Signals
from bitcoin_safe.signer import (
    AbstractSignatureImporter,
    SignatureImporterClipboard,
    SignatureImporterFile,
    SignatureImporterQR,
    SignatureImporterUSB,
    SignatureImporterWallet,
)
from bitcoin_safe.threading_manager import ThreadingManager

logger = logging.getLogger(__name__)
from PyQt6.QtWidgets import QWidget


class HorizontalImporters(HorizontalImportExportGroups):
    def __init__(
        self,
        signature_importers: List[AbstractSignatureImporter],
        psbt: bdk.PartiallySignedTransaction,
        network: bdk.Network,
    ) -> None:
        super().__init__()
        self.signature_importers = signature_importers
        self.psbt = psbt
        self.network = network

        self._add(self.group_qr, SignatureImporterQR)
        self._add(self.group_file, SignatureImporterFile)
        self._add(self.group_usb, SignatureImporterUSB)
        self._add(self.group_share, SignatureImporterClipboard)
        self._add(self.group_seed, SignatureImporterWallet)

    def _add(self, group: DataGroupBox, cls: Type[AbstractSignatureImporter]) -> None:
        importer = self._get_importer(cls)
        group.setVisible(bool(importer))
        if importer:
            signerui = SignerUI(
                [importer],
                self.psbt,
                self.network,
            )
            group._layout.addWidget(signerui)
            group.setData(signerui)

    def _get_importer(self, cls: Type[AbstractSignatureImporter]) -> Optional[AbstractSignatureImporter]:
        for importer in self.signature_importers:
            if isinstance(importer, cls):
                return importer
        return None


class TxSigningSteps(StepProgressContainer):
    def __init__(
        self,
        signature_importer_dict: Dict[str, List[AbstractSignatureImporter]],
        psbt: bdk.PartiallySignedTransaction,
        network: bdk.Network,
        signals: Signals,
        parent: QWidget | None = None,
        threading_parent: ThreadingManager | None = None,
    ) -> None:
        step_labels = []
        self.sub_indices = []
        enumeration_alphabet = []
        for i, (wallet_id, signature_importer_list) in enumerate(signature_importer_dict.items()):
            # export
            step_labels.append(
                (
                    self.tr("Export transaction to any hardware signer")
                    if i == 0
                    else self.tr("Sign with a different hardware signer")
                )
            )
            self.sub_indices.append(self._get_idx(i, 0))
            enumeration_alphabet.append(self._get_name(i, 0))

            # import
            step_labels.append(self.tr("Import signature"))
            enumeration_alphabet.append(self._get_name(i, 1))

        super().__init__(
            step_labels=step_labels,
            parent=parent,
            use_resizing_stacked_widget=False,
            signals_min=signals,
            threading_parent=threading_parent,
        )
        self.step_bar.set_enumeration_alphabet(enumeration_alphabet)
        self.set_sub_indices(self.sub_indices)

        self.psbt = psbt
        self.network = network
        self.signals = signals
        self.signature_importer_dict = signature_importer_dict

        first_non_signed_index = None
        # fill ui
        for i, (wallet_id, signature_importer_list) in enumerate(signature_importer_dict.items()):
            self.set_custom_widget(self._get_idx(i, 0), self.create_export_widget(signature_importer_list))
            self.set_custom_widget(self._get_idx(i, 1), self.create_import_widget(signature_importer_list))
            # set the index, to the first unsigned step
            if first_non_signed_index is None and (not signature_importer_list[0].signature_available):
                first_non_signed_index = i
                self.set_current_index(self._get_idx(i, 0))

    def _get_idx(self, i: int, j: int) -> int:
        return 2 * i + j

    def _get_name(self, i: int, j: int) -> str:
        alphabet = "abcdefghijklmnopqrstuvwxyz"
        return f"{i+1}.{alphabet[j]}"

    def set_current_index(self, index: int) -> None:
        super().set_current_index(index)

    def go_to_next_index(self) -> None:
        if self.current_index() + 1 < self.count():
            self.set_current_index(self.current_index() + 1)
        else:
            pass
            # do not mark as completed,.. Only successful signing can do this
            # self.step_bar.set_mark_current_step_as_completed(True)

    def go_to_previous_index(self) -> None:
        self.step_bar.set_mark_current_step_as_completed(False)

        if self.current_index() - 1 >= 0:
            self.set_current_index(self.current_index() - 1)

    def create_export_widget(self, signature_importers: List[AbstractSignatureImporter]) -> QWidget:
        usb_signers = [
            signature_importer
            for signature_importer in signature_importers
            if isinstance(signature_importer, SignatureImporterUSB)
        ]
        usb_signer_ui = None
        if usb_signers:
            usb_signer_ui = SignerUI(
                usb_signers,
                self.psbt,
                self.network,
            )

        export_widget = ExportDataSimple(
            data=Data.from_psbt(self.psbt),
            sync_tabs={
                wallet_id: qt_wallet.sync_tab
                for wallet_id, qt_wallet in self.signals.get_qt_wallets().items()
            },
            usb_signer_ui=usb_signer_ui,
            signals_min=self.signals,
            network=self.network,
            threading_parent=self.threading_parent,
        )

        export_widget.qr_label.set_always_animate(True)

        return export_widget

    def create_import_widget(self, signature_importers: List[AbstractSignatureImporter]) -> QWidget:
        if signature_importers[0].signature_available:
            return SignedUI(
                self.tr("Transaction signed with the private key belonging to {label}").format(
                    label=signature_importers[0].key_label
                ),
                self.psbt,
                self.network,
            )
        else:
            return HorizontalImporters(
                signature_importers,
                self.psbt,
                self.network,
            )


if __name__ == "__main__":
    import sys

    from PyQt6.QtWidgets import QApplication

    app = QApplication(sys.argv)

    # dialog = TxSigningSteps(2, [], [])
    # dialog.show()

    sys.exit(app.exec())



#
# Bitcoin Safe
# Copyright (C) 2024 Andreas Griffin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/gpl-3.0.html
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import enum
import logging
import platform
import shlex
import subprocess
import sys
import traceback
import webbrowser
import xml.etree.ElementTree as ET
from functools import lru_cache
from io import BytesIO
from pathlib import Path
from typing import Any, Callable, Iterable, List, Optional, Tuple, Union
from urllib.parse import urlparse

import bdkpython as bdk
import PIL.Image as PilImage
from bitcoin_qr_tools.data import is_bitcoin_address
from PyQt6.QtCore import (
    QByteArray,
    QCoreApplication,
    QSize,
    Qt,
    QTimer,
    QUrl,
    pyqtBoundSignal,
)
from PyQt6.QtGui import (
    QColor,
    QCursor,
    QDesktopServices,
    QFontMetrics,
    QIcon,
    QImage,
    QPalette,
    QPixmap,
)
from PyQt6.QtSvgWidgets import QSvgWidget
from PyQt6.QtWidgets import (
    QApplication,
    QBoxLayout,
    QDialog,
    QDialogButtonBox,
    QFileDialog,
    QHBoxLayout,
    QLabel,
    QLayout,
    QLineEdit,
    QMessageBox,
    QPushButton,
    QSizePolicy,
    QStyle,
    QSystemTrayIcon,
    QToolTip,
    QVBoxLayout,
    QWidget,
)

from bitcoin_safe.gui.qt.custom_edits import AnalyzerState

from ...i18n import translate
from ...util import register_cache, resource_path

logger = logging.getLogger(__name__)


if platform.system() == "Windows":
    MONOSPACE_FONT = "Lucida Console"
elif platform.system() == "Darwin":
    MONOSPACE_FONT = "Monaco"
else:
    MONOSPACE_FONT = "monospace"

QWIDGETSIZE_MAX = 16777215


# filter tx files in QFileDialog:
TRANSACTION_FILE_EXTENSION_FILTER_ANY = translate("constant", "Transaction (*.txn *.psbt);;All files (*)")
TRANSACTION_FILE_EXTENSION_FILTER_ONLY_PARTIAL_TX = translate("constant", "Partial Transaction (*.psbt)")
TRANSACTION_FILE_EXTENSION_FILTER_ONLY_COMPLETE_TX = translate("constant", "Complete Transaction (*.txn)")
TRANSACTION_FILE_EXTENSION_FILTER_SEPARATE = (
    f"{TRANSACTION_FILE_EXTENSION_FILTER_ONLY_PARTIAL_TX};;"
    f"{TRANSACTION_FILE_EXTENSION_FILTER_ONLY_COMPLETE_TX};;" + translate("constant", "All files (*)")
)


TX_ICONS: List[str] = [
    "unconfirmed.svg",
    "clock1.png",
    "clock2.png",
    "clock3.png",
    "clock4.png",
    "clock5.png",
    "confirmed.svg",
]


class QtWalletBase(QWidget):
    pass


def xdg_open_file(filename: Path):
    system_name = platform.system()
    if system_name == "Windows":
        subprocess.call(shlex.split(f'start "" /max "{filename}"'), shell=True)
    elif system_name == "Darwin":  # macOS
        subprocess.call(shlex.split(f'open "{filename}"'))
    elif system_name == "Linux":  # Linux
        subprocess.call(shlex.split(f'xdg-open "{filename}"'))


def sort_id_to_icon(sort_id: int) -> str:
    if sort_id < 0:
        return "offline_tx.png"
    if sort_id > len(TX_ICONS) - 1:
        sort_id = len(TX_ICONS) - 1

    return TX_ICONS[sort_id]


def open_website(url: str):
    QDesktopServices.openUrl(QUrl(url))


def resize(x, y, x_max, y_max):
    def resize_one_side(a, b, amax):
        a_new = min(amax, a)
        return a_new, b * a_new / a

    # resize according to xmax
    x, y = resize_one_side(x, y, x_max)
    # resize according to ymax
    y, x = resize_one_side(y, x, y_max)
    return x, y


def qresize(qsize: QSize, max_sizes: Tuple[int, int]):
    x, y = resize(qsize.width(), qsize.height(), *max_sizes)
    return QSize(int(x), int(y))


def center_in_widget(
    widgets: Iterable[QWidget], parent: QWidget, direction="h", alignment=Qt.AlignmentFlag.AlignCenter
):
    outer_layout = QHBoxLayout(parent) if direction == "h" else QVBoxLayout(parent)
    outer_layout.setAlignment(alignment)
    for widget in widgets:
        outer_layout.addWidget(widget)
        # outer_layout.setAlignment(widget, alignment)
    return outer_layout


def generate_help_button(help_widget: QWidget, title=translate("help", "Help")) -> QPushButton:
    # add the help buttonbox
    button_help = QPushButton()
    button_help.setSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Fixed)
    button_help.setText(title)
    button_help.setIcon(
        (button_help.style() or QStyle()).standardIcon(QStyle.StandardPixmap.SP_MessageBoxQuestion)
    )

    def show_screenshot_tutorial():
        help_widget.setWindowTitle(title)
        help_widget.show()

    button_help.clicked.connect(show_screenshot_tutorial)
    return button_help


def generate_help_message_button(message, title=translate("help", "Help")) -> QPushButton:
    msg_box = QMessageBox()
    msg_box.setWindowTitle(title)
    msg_box.setIcon(QMessageBox.Icon.Information)  # Set icon to Information which is often used for Help
    msg_box.setText(message)
    msg_box.setStandardButtons(QMessageBox.StandardButton.Ok)  # Add OK button

    help_button = generate_help_button(msg_box, title=title)
    return help_button


class AspectRatioSvgWidget(QSvgWidget):
    def __init__(self, svg_path: str, max_width: int, max_height: int, parent=None):
        super().__init__(parent)
        self.svg_path = svg_path
        self.load(svg_path)
        self._max_width = max_width
        self._max_height = max_height
        # self.setMinimumSize(max_width, max_height)

        self.setSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Fixed)
        self.setFixedSize(self.calculate_proportional_size())

    def calculate_proportional_size(self):
        qsize = qresize(self.sizeHint(), (self._max_width, self._max_height))
        return qsize

    def modify_svg_text(self, old_text: str, new_text: str):

        # Load the original SVG content
        with open(self.svg_path, "r", encoding="utf-8") as file:
            original_svg_content = file.read()

        # Register namespaces if needed (for saving purposes)
        namespaces = {
            "": "http://www.w3.org/2000/svg",
            "ns0": "http://www.w3.org/2000/svg",
            "ns1": "http://www.w3.org/2000/svg",
        }

        # Parse the SVG content
        tree = ET.ElementTree(ET.fromstring(original_svg_content))
        root = tree.getroot()

        # Find all text elements in the SVG
        for tspan in root.findall(".//{http://www.w3.org/2000/svg}tspan"):
            if tspan.text == old_text:
                tspan.text = new_text

        # Convert the modified SVG tree back to a bytes representation
        fake_file = BytesIO()
        tree.write(fake_file, encoding="utf-8", xml_declaration=True)

        byte_data = fake_file.getvalue()
        modified_svg_content = QByteArray(byte_data)  # type: ignore[call-overload]
        self.load(modified_svg_content)


def add_centered_icons(
    paths: List[str],
    parent_layout: QBoxLayout,
    direction="h",
    alignment: Qt.AlignmentFlag = Qt.AlignmentFlag.AlignCenter,
    max_sizes: Iterable[Tuple[int, int]] = [],
):
    max_sizes = max_sizes if max_sizes else [(60, 80) for path in paths]
    if len(paths) > 1 and len(max_sizes) == 1:  # type: ignore
        max_sizes = max_sizes * len(paths)  # type: ignore

    svg_widgets = [
        AspectRatioSvgWidget(icon_path(path), *max_size) for max_size, path in zip(max_sizes, paths)
    ]

    widget1 = QWidget()
    parent_layout.addWidget(widget1)
    inner_layout = center_in_widget(svg_widgets, widget1, direction=direction, alignment=alignment)
    inner_layout.setContentsMargins(1, 0, 1, 0)  # left, top, right, bottom

    return svg_widgets


def add_to_buttonbox(
    buttonBox: QDialogButtonBox,
    text: str,
    icon_name: str | None = None,
    on_clicked=None,
    role=QDialogButtonBox.ButtonRole.ActionRole,
):
    # Create a custom QPushButton with an icon
    button = QPushButton(text)
    if icon_name:
        button.setIcon(QIcon(icon_path(icon_name)))

    # Add the button to the QDialogButtonBox
    buttonBox.addButton(button, role)

    # Optionally connect the button's clicked signal
    if on_clicked:
        button.clicked.connect(on_clicked)
    return button


class Buttons(QHBoxLayout):
    def __init__(self, *buttons):
        QHBoxLayout.__init__(self)
        self.addStretch(1)
        for b in buttons:
            if b is None:
                continue
            self.addWidget(b)


class CloseButton(QPushButton):
    def __init__(self, dialog):
        QPushButton.__init__(self, self.tr("Close"))
        self.clicked.connect(dialog.close)
        self.setDefault(True)


class MessageType(enum.Enum):
    Info = enum.auto()
    Warning = enum.auto()
    Error = enum.auto()
    Critical = enum.auto()

    @classmethod
    def from_analyzer_state(cls, analyzer_state: AnalyzerState) -> "MessageType":
        return list(MessageType)[int(analyzer_state) - 1]


class Message:
    def __init__(
        self,
        msg: str,
        parent: QWidget | None = None,
        title: str | None = None,
        icon: Union[QIcon, QPixmap, QMessageBox.Icon] | None = None,
        msecs=None,
        type: MessageType = MessageType.Info,
        no_show=False,
        **kwargs,
    ) -> None:

        self.msg = msg
        self.parent = parent
        self.title = title
        self.icon = icon
        self.msecs = msecs
        self.type = type
        self.kwargs = kwargs

        if not no_show:
            self.show()

    @staticmethod
    def system_tray_icon(
        icon: Optional[Union[QIcon, QPixmap, QMessageBox.Icon, QSystemTrayIcon.MessageIcon]]
    ) -> Union[QIcon, QSystemTrayIcon.MessageIcon]:
        if isinstance(icon, QIcon):
            return icon

        if isinstance(icon, QSystemTrayIcon.MessageIcon):
            return icon

        if isinstance(icon, QPixmap):
            return QIcon(icon)

        if type(icon) == QMessageBox.Icon:
            if icon == QMessageBox.Icon.Information:
                return QSystemTrayIcon.MessageIcon.Information
            if icon == QMessageBox.Icon.Warning:
                return QSystemTrayIcon.MessageIcon.Warning
            if icon == QMessageBox.Icon.Critical:
                return QSystemTrayIcon.MessageIcon.Critical

        return QSystemTrayIcon.MessageIcon.NoIcon

    def get_icon_and_title(self) -> Tuple[Union[QIcon, QPixmap, QMessageBox.Icon], str]:
        icon = QMessageBox.Icon.Information
        title = "Information"
        if self.type in [MessageType.Warning]:
            icon = QMessageBox.Icon.Warning
            title = "Warning"
        if self.type in [MessageType.Error]:
            icon = QMessageBox.Icon.Warning
            title = "Error"
        if self.type in [MessageType.Critical]:
            icon = QMessageBox.Icon.Critical
            title = "Critical Error"

        return_icon = self.icon or icon
        title = self.title or title
        return return_icon, title

    def show(self):
        self.create().exec()

    def create(self) -> QMessageBox:
        logger.warning(str(self.__dict__))

        icon, title = self.get_icon_and_title()

        return self.msg_box(
            icon,
            self.parent,
            title,
            self.msg,
            **self.kwargs,
        )

    def ask(
        self,
        yes_button: QMessageBox.StandardButton = QMessageBox.StandardButton.Ok,
        no_button: QMessageBox.StandardButton = QMessageBox.StandardButton.Cancel,
    ) -> bool:
        msg_box = self.create()
        msg_box.setStandardButtons(yes_button | no_button)
        ret = msg_box.exec()

        # Check which button was clicked
        if ret == yes_button:
            return True
        elif ret == no_button:
            return False
        return False

    def emit_with(self, notification_signal: pyqtBoundSignal):
        logger.debug(str(self.__dict__))
        return notification_signal.emit(self)

    def msg_box(
        self,
        icon: Union[QIcon, QPixmap, QMessageBox.Icon],
        parent,
        title: str,
        text: str,
        *,
        buttons=QMessageBox.StandardButton.Ok,
        defaultButton=QMessageBox.StandardButton.NoButton,
        rich_text=True,
        checkbox=None,
    ) -> QMessageBox:
        # parent = parent or self.top_level_window()
        return custom_message_box(
            icon=icon,
            parent=parent,
            title=title,
            text=text,
            buttons=buttons,
            defaultButton=defaultButton,
            rich_text=rich_text,
            checkbox=checkbox,
        )


def custom_exception_handler(exc_type, exc_value, exc_traceback=None):
    """Custom exception handler to catch unhandled exceptions and display an
    error message box."""
    title = "Error"
    try:
        # Format the traceback for the email
        "".join(traceback.format_exception(exc_type, exc_value, exc_traceback))
        error_message = f"{exc_type.__name__}: {exc_value}"

        logger.critical(error_message, exc_info=(exc_type, exc_value, exc_traceback))
        QMessageBox.critical(
            None,  # type: ignore[arg-type]
            title,
            f"{error_message}\n\nPlease send the error report via email, so the bug can be fixed.",
        )

    except:
        error_message = str([exc_type, exc_value, exc_traceback])
        logger.critical(error_message)
        QMessageBox.critical(
            None,  # type: ignore[arg-type]
            title,
            f"{error_message}\n\nPlease send the error report via email, so the bug can be fixed.",
        )


def caught_exception_message(e: Exception, title=None, log_traceback=False) -> Message:
    exception_text = str(e).replace("\\", "")

    logger.error(exception_text, exc_info=sys.exc_info() if log_traceback else None)

    text = title + "\n\n" if title else ""
    text += exception_text
    return Message(text, type=MessageType.Error)


def custom_message_box(
    *,
    icon: Union[QIcon, QPixmap, QMessageBox.Icon],
    parent,
    title: str,
    text: str,
    buttons=QMessageBox.StandardButton.Ok,
    defaultButton=QMessageBox.StandardButton.NoButton,
    rich_text=False,
    checkbox=None,
) -> QMessageBox:

    if not isinstance(icon, (QIcon, QPixmap, QMessageBox.Icon)):
        raise ValueError(f"{icon} is not a valid type")

    if isinstance(icon, QMessageBox.Icon):
        d = QMessageBox(icon, title, str(text), buttons, parent)
    else:
        d = QMessageBox(QMessageBox.Icon.Information, title, str(text), buttons, parent)
        pixmap_icon = None
        if isinstance(icon, QPixmap):
            pixmap_icon = icon
        if isinstance(icon, QIcon):
            pixmap_icon = icon.pixmap(60, 60)

        if pixmap_icon:
            d.setIconPixmap(pixmap_icon)

    d.setWindowModality(Qt.WindowModality.WindowModal)
    d.setDefaultButton(defaultButton)
    if rich_text:
        d.setTextInteractionFlags(
            Qt.TextInteractionFlag.TextSelectableByMouse | Qt.TextInteractionFlag.LinksAccessibleByMouse
        )
        # set AutoText instead of RichText
        # AutoText lets Qt figure out whether to render as rich text.
        # e.g. if text is actually plain text and uses "\n" newlines;
        #      and we set RichText here, newlines would be swallowed
        d.setTextFormat(Qt.TextFormat.AutoText)
    else:
        d.setTextInteractionFlags(Qt.TextInteractionFlag.TextSelectableByMouse)
        d.setTextFormat(Qt.TextFormat.PlainText)
    if checkbox is not None:
        d.setCheckBox(checkbox)
    return d


class WindowModalDialog(QDialog):
    """Handy wrapper; window modal dialogs are better for our multi-window
    daemon model as other wallet windows can still be accessed."""

    def __init__(self, parent, title=None):
        QDialog.__init__(self, parent)
        self.setWindowModality(Qt.WindowModality.WindowModal)
        if title:
            self.setWindowTitle(title)


class BlockingWaitingDialog(WindowModalDialog):
    """Shows a waiting dialog whilst running a task.

    Should be called from the GUI thread. The GUI thread will be blocked
    while the task is running; the point of the dialog is to provide
    feedback to the user regarding what is going on.
    """

    def __init__(self, parent: QWidget, message: str, task: Callable[[], Any]):
        assert parent
        WindowModalDialog.__init__(self, parent, self.tr("Please wait"))
        self.message_label = QLabel(message)
        vbox = QVBoxLayout(self)
        vbox.addWidget(self.message_label)
        self.finished.connect(self.deleteLater)  # see #3956
        # show popup
        self.show()
        # refresh GUI; needed for popup to appear and for message_label to get drawn
        QCoreApplication.processEvents()
        QCoreApplication.processEvents()
        try:
            # block and run given task
            task()
        finally:
            # close popup
            self.accept()


def one_time_signal_connection(signal: pyqtBoundSignal, f: Callable):
    def f_wrapper(*args, **kwargs):
        signal.disconnect(f_wrapper)
        return f(*args, **kwargs)

    signal.connect(f_wrapper)


def chained_one_time_signal_connections(
    signals: List[pyqtBoundSignal], fs: List[Callable[..., bool]], disconnect_only_if_f_true=True
):
    "If after the i. f is called, it connects the i+1. signal"

    signal, remaining_signals = signals[0], signals[1:]
    f, remaining_fs = fs[0], fs[1:]

    def f_wrapper(*args, **kwargs):
        res = f(*args, **kwargs)
        if disconnect_only_if_f_true and not res:
            # reconnect
            one_time_signal_connection(signal, f_wrapper)
        elif remaining_signals and remaining_fs:
            chained_one_time_signal_connections(remaining_signals, remaining_fs)
        return res

    one_time_signal_connection(signal, f_wrapper)


def create_button_box(
    callback_ok, callback_cancel, ok_text=None, cancel_text=None
) -> Tuple[QDialogButtonBox, List[QPushButton]]:
    # Create the QDialogButtonBox instance
    button_box = QDialogButtonBox()
    buttons: List[QPushButton] = []

    # Add an 'Ok' button
    if ok_text is None:
        ok_button = button_box.addButton(QDialogButtonBox.StandardButton.Ok)
        if ok_button:
            buttons.append(ok_button)
    else:
        custom_yes_button = QPushButton(ok_text)
        buttons.append(custom_yes_button)
        button_box.addButton(custom_yes_button, QDialogButtonBox.ButtonRole.AcceptRole)
        custom_yes_button.clicked.connect(callback_ok)

    # Add a 'Cancel' button
    if cancel_text is None:
        cancel_button = button_box.addButton(QDialogButtonBox.StandardButton.Cancel)
        if cancel_button:
            buttons.append(cancel_button)
    else:
        custom_cancel_button = QPushButton(cancel_text)
        buttons.append(custom_cancel_button)
        button_box.addButton(custom_cancel_button, QDialogButtonBox.ButtonRole.RejectRole)
        custom_cancel_button.clicked.connect(callback_cancel)

    # Connect the QDialogButtonBox's accepted and rejected signals if default buttons are used
    if ok_text is None:
        button_box.accepted.connect(callback_ok)
    if cancel_text is None:
        button_box.rejected.connect(callback_cancel)

    return button_box, buttons


class ColorSchemeItem:
    def __init__(self, fg_color, bg_color):
        self.colors = (fg_color, bg_color)

    @register_cache(always_keep=True)
    def _get_color(self, background):
        return self.colors[(int(background) + int(ColorScheme.dark_scheme)) % 2]

    def as_stylesheet(self, background=False):
        css_prefix = "background-" if background else ""
        color = self._get_color(background)
        return "QWidget {{ {}color:{}; }}".format(css_prefix, color)

    def as_color(self, background=False):
        color = self._get_color(background)
        return QColor(color)


class ColorScheme:
    dark_scheme = False

    GREEN = ColorSchemeItem("#117c11", "#8af296")
    YELLOW = ColorSchemeItem("#897b2a", "#ffff00")
    RED = ColorSchemeItem("#7c1111", "#f18c8c")
    BLUE = ColorSchemeItem("#123b7c", "#8cb3f2")
    DEFAULT = ColorSchemeItem("black", "white")
    GRAY = ColorSchemeItem("gray", "gray")

    @staticmethod
    def has_dark_background(widget: QWidget):
        background_color = widget.palette().color(QPalette.ColorGroup.Normal, QPalette.ColorRole.Window)
        rgb = background_color.getRgb()[0:3]
        brightness = sum([c for c in rgb if c])
        return brightness < (255 * 3 / 2)

    @staticmethod
    def update_from_widget(widget, force_dark=False):
        ColorScheme.dark_scheme = bool(force_dark or ColorScheme.has_dark_background(widget))


def icon_path(icon_basename: str):
    return resource_path("gui", "icons", icon_basename)


def screenshot_path(basename: str):
    return resource_path("gui", "screenshots", basename)


@lru_cache(maxsize=1000)
def read_QIcon(icon_basename: Optional[str]) -> QIcon:
    if not icon_basename:
        return QIcon()
    return QIcon(icon_path(icon_basename))


def char_width_in_lineedit() -> int:
    char_width = QFontMetrics(QLineEdit().font()).averageCharWidth()
    # 'averageCharWidth' seems to underestimate on Windows, hence 'max()'
    return max(9, char_width)


def font_height() -> int:
    return QFontMetrics(QLabel().font()).height()


def webopen(url: str):
    webbrowser.open(url)


def clipboard_contains_address(network: bdk.Network) -> bool:
    clipboard = QApplication.clipboard()
    if not clipboard:
        return False
    return is_bitcoin_address(clipboard.text(), network)


def do_copy(text: str, *, title: str | None = None) -> None:
    clipboard = QApplication.clipboard()
    if not clipboard:
        show_tooltip_after_delay("Clipboard not available")
        return
    clipboard.setText(str(text))
    message = (
        translate("d", "Text copied to Clipboard")
        if title is None
        else translate("d", "{} copied to Clipboard").format(title)
    )
    show_tooltip_after_delay(message)


def show_tooltip_after_delay(message):
    timer = QTimer()
    # tooltip cannot be displayed immediately when called from a menu; wait 200ms
    timer.singleShot(200, lambda: QToolTip.showText(QCursor.pos(), message))


def qicon_to_pil(qicon: QIcon, size=200) -> PilImage.Image:
    # Convert QIcon to QPixmap
    pixmap = qicon.pixmap(size, size)  # specify the size you want

    # Convert QPixmap to QImage
    qimage = pixmap.toImage().convertToFormat(QImage.Format.Format_RGBA8888)

    # Get image dimensions
    img_size = qimage.size()
    width, height = img_size.width(), img_size.height()

    # Convert QImage to raw bytes
    buffer = qimage.bits()
    if buffer:
        buffer.setsize(width * height * 4)  # RGBA8888 = 4 bytes per pixel

    # Convert raw bytes to a PIL image
    pil_image = PilImage.frombuffer("RGBA", (width, height), bytes(buffer), "raw", "RGBA", 0, 1)  # type: ignore[call-overload]

    return pil_image


def save_file_dialog(name_filters=None, default_suffix=None, default_filename=None) -> Optional[str]:
    file_dialog = QFileDialog()
    file_dialog.setWindowTitle("Save File")
    if default_suffix:
        file_dialog.setDefaultSuffix(default_suffix)

    # Set a default filename
    if default_filename:
        file_dialog.selectFile(default_filename)

    file_dialog.setAcceptMode(QFileDialog.AcceptMode.AcceptSave)
    if name_filters:
        file_dialog.setNameFilters(name_filters)

    if file_dialog.exec() == QFileDialog.DialogCode.Accepted:
        selected_file = file_dialog.selectedFiles()[0]
        # Do something with the selected file path, e.g., save data to the file
        logger.debug(f"Selected save file: {selected_file}")
        return selected_file
    return None


def remove_scheme(url):
    """Check if "://" is in the URL and split it"""
    if "://" in url:
        parts = url.split("://", 1)  # Split only at the first occurrence
        return parts[1]  # Return the part after the "://"
    else:
        return url  # Return the original URL if no scheme is found


def ensure_scheme(url, default_scheme="https://"):
    """Check if "://" is in the URL and split it"""
    if "://" in url:
        return url  # Return the original URL if   scheme is found
    else:
        return f"{default_scheme}{url}"


def get_host_and_port(url) -> Tuple[str | None, int | None]:

    parsed_url = urlparse(ensure_scheme(url))

    # Extract the hostname and port
    return parsed_url.hostname, parsed_url.port


def delayed_execution(f, parent, delay=10):
    timer = QTimer(parent)
    timer.setSingleShot(True)  # Make sure the timer runs only once
    timer.timeout.connect(f)  # Connect the timeout signal to the function
    timer.start(delay)


def clear_layout(layout: QLayout) -> None:
    """Helper method to remove all widgets from the grid layout."""
    while layout.count():
        item = layout.takeAt(0)
        if not item:
            continue
        widget = item.widget()
        if widget:
            layout.removeWidget(widget)
            widget.setParent(None)  # Remove widget from parent to fully disconnect it
            widget.deleteLater()



#
# Bitcoin Safe
# Copyright (C) 2024 Andreas Griffin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/gpl-3.0.html
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import logging
from typing import Iterable, Optional, Tuple, Union

from bitcoin_safe.gui.qt.analyzer_indicator import AnalyzerIndicator
from bitcoin_safe.gui.qt.analyzers import (
    FingerprintAnalyzer,
    KeyOriginAnalyzer,
    SeedAnalyzer,
    XpubAnalyzer,
)
from bitcoin_safe.gui.qt.data_tab_widget import DataTabWidget
from bitcoin_safe.gui.qt.wrappers import Menu
from bitcoin_safe.i18n import translate

from ...dynamic_lib_load import setup_libsecp256k1

setup_libsecp256k1()

from bitcoin_usb.address_types import SimplePubKeyProvider

from bitcoin_safe.gui.qt.buttonedit import ButtonEdit
from bitcoin_safe.gui.qt.custom_edits import AnalyzerTextEdit, QCompleterLineEdit
from bitcoin_safe.gui.qt.tutorial_screenshots import ScreenshotsExportXpub

from .dialog_import import ImportDialog

logger = logging.getLogger(__name__)

from typing import Callable, List

import bdkpython as bdk
from bitcoin_qr_tools.data import (
    Data,
    DataType,
    SignerInfo,
    convert_slip132_to_bip32,
    is_slip132,
)
from bitcoin_usb.address_types import AddressType
from bitcoin_usb.gui import USBGui
from bitcoin_usb.seed_tools import get_network_index
from bitcoin_usb.software_signer import SoftwareSigner
from PyQt6.QtCore import QObject, QSize, Qt, pyqtSignal
from PyQt6.QtGui import QIcon
from PyQt6.QtWidgets import (
    QDialogButtonBox,
    QFormLayout,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QPushButton,
    QSizePolicy,
    QTabWidget,
    QTextEdit,
    QToolButton,
    QVBoxLayout,
    QWidget,
)

from ...keystore import KeyStore, KeyStoreImporterTypes
from ...signals import SignalsMin, pyqtSignal
from ...signer import AbstractSignatureImporter
from .block_change_signals import BlockChangesSignals
from .util import (
    Message,
    MessageType,
    add_to_buttonbox,
    generate_help_button,
    icon_path,
    read_QIcon,
)


def icon_for_label(label: str) -> QIcon:
    return (
        read_QIcon("key-gray.png") if label.startswith(translate("d", "Recovery")) else read_QIcon("key.png")
    )


class HardwareSignerInteractionWidget(QWidget):
    def __init__(self) -> None:
        super().__init__()
        self._layout = QVBoxLayout(self)
        self._layout.setContentsMargins(0, 0, 0, 0)  # Left, Top, Right, Bottom margins

        # add the buttons
        self.buttonBox = QDialogButtonBox()
        self.button_import_file: Optional[QPushButton] = None
        self.button_import_qr: Optional[QPushButton] = None
        self.button_export_qr: Optional[QToolButton] = None
        self.button_hwi: Optional[QPushButton] = None
        self.help_button: Optional[QPushButton] = None
        self.button_export_file: Optional[QPushButton] = None

        self._layout.addWidget(self.buttonBox)
        self._layout.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self._layout.setAlignment(self.buttonBox, Qt.AlignmentFlag.AlignCenter)

    def add_import_file_button(self) -> QPushButton:

        # Create custom buttons
        self.button_import_file = button_import_file = add_to_buttonbox(
            self.buttonBox, self.tr(""), KeyStoreImporterTypes.file.icon_filename
        )
        return button_import_file

    def add_export_file_button(self) -> QPushButton:

        # Create custom buttons
        self.button_export_file = button_export_file = add_to_buttonbox(
            self.buttonBox, self.tr(""), KeyStoreImporterTypes.file.icon_filename
        )
        return button_export_file

    def add_qr_import_buttonn(self) -> QPushButton:
        self.button_import_qr = button_import_qr = add_to_buttonbox(
            self.buttonBox, (""), KeyStoreImporterTypes.qr.icon_filename
        )
        return button_import_qr

    def add_export_qr_button(self) -> Tuple[QToolButton, Menu]:

        # Create a custom QPushButton with an icon
        button = QToolButton(self)
        button.setIcon(QIcon(icon_path(KeyStoreImporterTypes.qr.icon_filename)))
        button.setToolButtonStyle(Qt.ToolButtonStyle.ToolButtonTextBesideIcon)

        # Add the button to the QDialogButtonBox
        self.buttonBox.addButton(button, QDialogButtonBox.ButtonRole.ActionRole)

        menu = Menu(self)
        button.setMenu(menu)
        button.setPopupMode(QToolButton.ToolButtonPopupMode.InstantPopup)

        self.button_export_qr = button
        return self.button_export_qr, menu

    def add_hwi_button(self) -> QPushButton:
        button_hwi = add_to_buttonbox(self.buttonBox, (""), KeyStoreImporterTypes.hwi.icon_filename)
        self.button_hwi = button_hwi
        return button_hwi

    def add_help_button(self, help_widget: QWidget) -> QPushButton:
        self.buttonBoxHelp = QDialogButtonBox()
        help_button = generate_help_button(help_widget)

        self.buttonBoxHelp.addButton(help_button, QDialogButtonBox.ButtonRole.ActionRole)
        self._layout.addWidget(self.buttonBoxHelp)
        self._layout.setAlignment(self.buttonBoxHelp, Qt.AlignmentFlag.AlignCenter)

        self.help_button = help_button
        return help_button

    def updateUi(self) -> None:
        if self.button_import_file:
            self.button_import_file.setText(self.tr("Import File or Text"))
        if self.button_export_file:
            self.button_export_file.setText(self.tr("Export File"))
        if self.button_import_qr:
            self.button_import_qr.setText(self.tr("QR Code"))
        if self.button_export_qr:
            self.button_export_qr.setText(self.tr("QR Code"))
        if self.button_hwi:
            self.button_hwi.setText(self.tr("USB"))
        if self.help_button:
            self.help_button.setText(self.tr("Help"))


class KeyStoreUI(QObject):
    signal_signer_infos = pyqtSignal(list)

    def __init__(
        self,
        tabs: DataTabWidget,
        network: bdk.Network,
        get_address_type: Callable[[], AddressType],
        signals_min: SignalsMin,
        label: str = "",
        hardware_signer_label="",
    ) -> None:
        super().__init__()
        self.signals_min = signals_min

        self.label = label
        self.hardware_signer_label = hardware_signer_label
        self.tabs = tabs
        self.network = network
        self.get_address_type = get_address_type

        self.tab = QWidget()
        self.tab_layout = QHBoxLayout(self.tab)

        self.tabs_import_type = QTabWidget()
        self.tab_layout.addWidget(self.tabs_import_type)

        self.tab_import = QWidget()
        self.tab_import_layout = QVBoxLayout(self.tab_import)
        self.tabs_import_type.addTab(self.tab_import, "")
        self.tab_manual = QWidget()
        self.tabs_import_type.addTab(self.tab_manual, "")

        self.label_fingerprint = QLabel()
        self.edit_fingerprint = ButtonEdit(
            signal_update=self.signals_min.language_switch,
        )
        self.edit_fingerprint.add_qr_input_from_camera_button(
            network=self.network,
        )
        self.edit_fingerprint.signal_data.connect(self._on_handle_input)

        self.edit_fingerprint.input_field.setAnalyzer(FingerprintAnalyzer(parent=self))
        # key_origin
        self.label_key_origin = QLabel()
        self.edit_key_origin_input = QCompleterLineEdit(self.network)
        self.edit_key_origin = ButtonEdit(
            input_field=self.edit_key_origin_input,
            signal_update=self.signals_min.language_switch,
        )
        self.edit_key_origin.add_qr_input_from_camera_button(
            network=self.network,
        )
        self.edit_key_origin.signal_data.connect(self._on_handle_input)
        self.edit_key_origin_input.setAnalyzer(
            KeyOriginAnalyzer(get_expected_key_origin=self.get_expected_key_origin, parent=self)
        )

        # xpub
        self.label_xpub = QLabel()
        self.edit_xpub = ButtonEdit(
            input_field=AnalyzerTextEdit(),
            signal_update=self.signals_min.language_switch,
        )
        self.edit_xpub.setFixedHeight(50)
        self.edit_xpub.add_qr_input_from_camera_button(
            network=self.network,
        )
        self.edit_xpub.signal_data.connect(self._on_handle_input)

        self.edit_xpub.input_field.setAnalyzer(XpubAnalyzer(self.network, parent=self))
        self.label_seed = QLabel()
        self.edit_seed = ButtonEdit()

        def callback_seed(seed: str) -> None:
            try:
                keystore = self.get_ui_values_as_keystore()
            except Exception as e:
                Message(str(e), type=MessageType.Error)
                return
            self.edit_fingerprint.setText(keystore.fingerprint)
            self.edit_xpub.setText(keystore.xpub)
            self.key_origin = keystore.key_origin

        self.edit_seed.add_random_mnemonic_button(callback_seed=callback_seed)
        self.edit_seed.input_field.setAnalyzer(SeedAnalyzer(parent=self))

        # put them on the formLayout
        self.formLayout = QFormLayout(self.tab_manual)
        self.formLayout.setFieldGrowthPolicy(QFormLayout.FieldGrowthPolicy.AllNonFixedFieldsGrow)
        self.formLayout.setWidget(1, QFormLayout.ItemRole.LabelRole, self.label_fingerprint)
        self.formLayout.setWidget(1, QFormLayout.ItemRole.FieldRole, self.edit_fingerprint)
        self.formLayout.setWidget(2, QFormLayout.ItemRole.LabelRole, self.label_key_origin)
        self.formLayout.setWidget(2, QFormLayout.ItemRole.FieldRole, self.edit_key_origin)
        self.formLayout.setWidget(3, QFormLayout.ItemRole.LabelRole, self.label_xpub)
        self.formLayout.setWidget(3, QFormLayout.ItemRole.FieldRole, self.edit_xpub)
        self.formLayout.setWidget(4, QFormLayout.ItemRole.LabelRole, self.label_seed)
        self.formLayout.setWidget(5, QFormLayout.ItemRole.FieldRole, self.edit_seed)
        self.seed_visibility(self.network in KeyStoreImporterTypes.seed.networks)

        # tab_import

        self.hardware_signer_interaction = HardwareSignerInteractionWidget()
        button_file = self.hardware_signer_interaction.add_import_file_button()
        button_qr = self.hardware_signer_interaction.add_qr_import_buttonn()
        self.hardware_signer_interaction.add_help_button(ScreenshotsExportXpub())

        button_qr.clicked.connect(lambda: self.edit_xpub.button_container.buttons[0].click())

        button_hwi = self.hardware_signer_interaction.add_hwi_button()
        button_hwi.clicked.connect(lambda: self.on_hwi_click())

        def process_input(s: str) -> None:
            res = Data.from_str(s, self.network)
            self._on_handle_input(res)

        def import_dialog():
            ImportDialog(
                self.network,
                on_open=process_input,
                window_title=self.tr("Import fingerprint and xpub"),
                text_button_ok=self.tr("OK"),
                text_instruction_label=self.tr("Please paste the exported file (like sparrow-export.json):"),
                text_placeholder=self.tr("Please paste the exported file (like sparrow-export.json)"),
            ).exec()

        button_file.clicked.connect(import_dialog)

        # self.tab_import_layout.addItem(QSpacerItem(1, 1, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum))
        self.tab_import_layout.setAlignment(Qt.AlignmentFlag.AlignCenter)

        icon_width = 50
        self.analyzer_indicator = AnalyzerIndicator(
            line_edits=[
                self.edit_fingerprint.input_field,
                self.edit_key_origin_input,
                self.edit_xpub.input_field,
            ],
            icon_OK=read_QIcon("checkmark.svg").pixmap(QSize(icon_width, icon_width)),
            icon_warning=read_QIcon("warning.png").pixmap(QSize(icon_width, icon_width)),
            icon_error=read_QIcon("error.png").pixmap(QSize(icon_width, icon_width)),
            hide_if_all_empty=True,
        )
        self.analyzer_indicator.setSizePolicy(QSizePolicy.Policy.Ignored, QSizePolicy.Policy.Preferred)
        self.tab_import_layout.addWidget(self.analyzer_indicator)
        self.tab_import_layout.addWidget(self.hardware_signer_interaction)

        # self.tab_import_layout.addItem(QSpacerItem(1, 1, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum))

        self.right_widget = QWidget()
        self.right_widget_layout = QVBoxLayout(self.right_widget)
        # self.right_widget_layout.setContentsMargins(0,0,0,0)

        self.label_description = QLabel()

        self.right_widget_layout.addWidget(self.label_description)

        self.textEdit_description = AnalyzerTextEdit()
        self.right_widget_layout.addWidget(self.textEdit_description)

        self.tab_layout.addWidget(self.right_widget)

        self.updateUi()

        self.edit_key_origin_input.textChanged.connect(self.format_all_fields)
        self.signals_min.language_switch.connect(self.updateUi)

    def seed_visibility(self, visible=False) -> None:

        self.edit_seed.setHidden(not visible)
        self.label_seed.setHidden(not visible)

        # self.edit_xpub.setHidden(visible)
        # self.edit_fingerprint.setHidden(visible)
        # self.label_xpub.setHidden(visible)
        # self.label_fingerprint.setHidden(visible)

    @property
    def key_origin(self) -> str:
        try:
            standardized = SimplePubKeyProvider.format_key_origin(self.edit_key_origin.text().strip())
        except:
            return ""

        return standardized

    @key_origin.setter
    def key_origin(self, value: str) -> None:
        self.edit_key_origin.setText(value if value else "")

    def format_all_fields(self) -> None:
        self.edit_fingerprint.format_and_apply_validator()

        expected_key_origin = self.get_expected_key_origin()
        if expected_key_origin != self.key_origin:
            self.edit_key_origin.format_as_error(True)
            self.edit_key_origin.setToolTip(
                self.tr(
                    "Standart for the selected address type {type} is {expected_key_origin}.  Please correct if you are not sure."
                ).format(expected_key_origin=expected_key_origin, type=self.get_address_type().name)
            )
            self.edit_xpub.format_as_error(True)
            self.edit_xpub.setToolTip(
                self.tr(
                    "The xPub origin {key_origin} and the xPub belong together. Please choose the correct xPub origin pair."
                ).format(key_origin=self.key_origin)
            )
        else:
            self.edit_xpub.format_and_apply_validator()
            self.edit_xpub.setToolTip("")
            self.edit_key_origin.format_and_apply_validator()
            self.edit_key_origin.setToolTip(f"")
        self.edit_key_origin.setPlaceholderText(expected_key_origin)
        self.edit_key_origin_input.reset_memory()
        self.edit_key_origin_input.add_to_memory(expected_key_origin)
        self.analyzer_indicator.updateUi()

    def get_expected_key_origin(self) -> str:
        return self.get_address_type().key_origin(self.network)

    def set_using_signer_info(self, signer_info: SignerInfo) -> None:
        def check_key_origin(signer_info: SignerInfo) -> bool:
            expected_key_origin = self.get_expected_key_origin()
            if signer_info.key_origin != expected_key_origin:
                if get_network_index(signer_info.key_origin) != get_network_index(expected_key_origin):
                    Message(
                        self.tr(
                            "The provided information is for {key_origin_network}. Please provide xPub for network {network}"
                        ).format(
                            key_origin_network=(
                                bdk.Network.BITCOIN
                                if get_network_index(signer_info.key_origin) == 1
                                else bdk.Network.REGTEST
                            ),
                            network=self.network,
                        ),
                        type=MessageType.Error,
                    )
                else:
                    Message(
                        self.tr(
                            "The xPub Origin {key_origin} is not the expected {expected_key_origin} for {address_type}"
                        ).format(
                            key_origin=signer_info.key_origin,
                            expected_key_origin=expected_key_origin,
                            address_type=self.get_address_type().name,
                        ),
                        type=MessageType.Error,
                    )
                return False
            return True

        if not check_key_origin(signer_info):
            return
        self.edit_xpub.setText(signer_info.xpub)
        self.key_origin = signer_info.key_origin
        self.edit_fingerprint.setText(signer_info.fingerprint)

    def _on_handle_input(self, data: Data, parent: Union[QLineEdit, QTextEdit] | None = None) -> None:

        if data.data_type == DataType.SignerInfo:
            self.set_using_signer_info(data.data)
        elif data.data_type == DataType.SignerInfos:
            expected_key_origin = self.get_expected_key_origin()

            matching_signer_infos = [
                signer_info
                for signer_info in data.data
                if isinstance(signer_info, SignerInfo) and (signer_info.key_origin == expected_key_origin)
            ]

            if len(matching_signer_infos) == 1:
                self.set_using_signer_info(matching_signer_infos[0])
            elif len(matching_signer_infos) > 1:
                self.signal_signer_infos.emit(matching_signer_infos)
            else:
                # none found
                Message(
                    self.tr("No signer data for the expected key_origin {expected_key_origin} found.").format(
                        expected_key_origin=expected_key_origin
                    )
                )

        elif data.data_type == DataType.Xpub:
            self.edit_xpub.setText(data.data)
        elif data.data_type == DataType.Fingerprint:
            self.edit_fingerprint.setText(data.data)
        elif data.data_type in [
            DataType.Descriptor,
            DataType.MultiPathDescriptor,
        ]:
            Message(self.tr("Please paste descriptors into the descriptor field in the top right."))
        elif isinstance(data.data, str) and parent:
            parent.setText(data.data)
        elif isinstance(data, Data):
            Message(
                self.tr("{data_type} cannot be used here.").format(data_type=data.data_type),
                type=MessageType.Error,
            )
        else:
            Exception("Could not recognize the QR Code")

    def xpub_validator(self) -> bool:
        xpub = self.edit_xpub.text()
        # automatically convert slip132
        if is_slip132(xpub):
            Message(
                self.tr("The xpub is in SLIP132 format. Converting to standard format."),
                title="Converting format",
            )
            try:
                self.edit_xpub.setText(convert_slip132_to_bip32(xpub))
            except:
                return False

        return KeyStore.is_xpub_valid(self.edit_xpub.text(), network=self.network)

    def updateUi(self) -> None:
        self.tabs.setTabText(
            self.tabs.indexOf(self.tab),
            self.label,
        )

        self.tabs_import_type.setTabText(self.tabs_import_type.indexOf(self.tab_import), self.tr("Import"))
        self.tabs_import_type.setTabText(self.tabs_import_type.indexOf(self.tab_manual), self.tr("Manual"))
        self.label_description.setText(self.tr("Description"))

        self.label_fingerprint.setText(self.tr("Fingerprint"))
        self.edit_fingerprint.input_field.setObjectName(self.label_fingerprint.text())

        self.label_key_origin.setText(self.tr("xPub Origin"))
        self.edit_key_origin_input.setObjectName(self.label_key_origin.text())

        self.label_xpub.setText(self.tr("xPub"))
        self.edit_xpub.input_field.setObjectName(self.label_xpub.text())

        self.label_seed.setText(self.tr("Seed"))
        self.edit_seed.input_field.setObjectName(self.label_seed.text())

        self.textEdit_description.setPlaceholderText(
            self.tr(
                "Name of signing device: ......\nLocation of signing device: .....",
            )
        )
        self.analyzer_indicator.updateUi()
        self.hardware_signer_interaction.updateUi()

    def on_hwi_click(self) -> None:
        address_type = self.get_address_type()
        usb = USBGui(self.network, initalization_label=self.hardware_signer_label)
        key_origin = address_type.key_origin(self.network)
        try:
            result = usb.get_fingerprint_and_xpub(key_origin=key_origin)
        except Exception as e:
            Message(
                str(e)
                + "\n\n"
                + self.tr("Please ensure that there are no other programs accessing the Hardware signer"),
                type=MessageType.Error,
            )
            return
        if not result:
            return

        device, fingerprint, xpub = result
        self.set_using_signer_info(SignerInfo(fingerprint=fingerprint, key_origin=key_origin, xpub=xpub))
        if not self.textEdit_description.text():
            self.textEdit_description.setText(f"{device.get('type', '')} - {device.get('model', '')}")

    def get_ui_values_as_keystore(self) -> KeyStore:
        seed_str = self.edit_seed.text().strip()

        if seed_str:
            mnemonic = bdk.Mnemonic.from_string(seed_str).as_string()
            software_signer = SoftwareSigner(mnemonic, self.network)
            key_origin = self.edit_key_origin.text().strip()
            # if key_origin is empty  fill it with the default
            key_origin = key_origin if key_origin else self.get_address_type().key_origin(self.network)
            xpub = software_signer.derive(key_origin)
            fingerprint = software_signer.get_fingerprint()
        else:
            mnemonic = None
            fingerprint = self.edit_fingerprint.text()
            xpub = self.edit_xpub.text()
            key_origin = self.key_origin

            # try to validate
            # if this works, then these are valid values
            if not KeyStore.is_xpub_valid(xpub, self.network):
                if xpub:
                    raise ValueError(self.tr("{xpub} is not a valid public xpub").format(xpub=xpub))
                else:
                    raise ValueError(self.tr("Please import the information from all hardware signers first"))

        return KeyStore(
            xpub=xpub,
            fingerprint=fingerprint,
            key_origin=key_origin,
            label=self.label,
            mnemonic=mnemonic if mnemonic else None,
            description=self.textEdit_description.toPlainText(),
            network=self.network,
        )

    def set_ui_from_keystore(self, keystore: KeyStore) -> None:
        with BlockChangesSignals([self.tab]):
            self.label = keystore.label

            logger.debug(f"{self.__class__.__name__} set_ui_from_keystore")
            xpub = keystore.xpub if keystore.xpub else ""
            if xpub != self.edit_xpub.text():
                self.edit_xpub.setText(xpub)

            fingerprint = keystore.fingerprint if keystore.fingerprint else ""
            if fingerprint != self.edit_fingerprint.text():
                self.edit_fingerprint.setText(fingerprint)

            if self.key_origin != keystore.key_origin:
                self.key_origin = keystore.key_origin

            if self.textEdit_description.toPlainText() != keystore.description:
                self.textEdit_description.setPlainText(keystore.description)

            mnemonic = keystore.mnemonic if keystore.mnemonic else ""
            if self.edit_seed.text() != mnemonic:
                self.edit_seed.setText(mnemonic)


class SignedUI(QWidget):
    def __init__(
        self,
        text: str,
        psbt: bdk.PartiallySignedTransaction,
        network: bdk.Network,
    ) -> None:
        super().__init__()
        self.text = text
        self.psbt = psbt
        self.network = network

        self.layout_keystore_buttons = QHBoxLayout(self)

        self.edit_signature = QTextEdit()
        self.edit_signature.setMinimumHeight(30)
        self.edit_signature.setReadOnly(True)
        self.edit_signature.setText(str(self.text))
        self.layout_keystore_buttons.addWidget(self.edit_signature)


class SignerUI(QWidget):
    signal_signature_added = pyqtSignal(bdk.PartiallySignedTransaction)
    signal_tx_received = pyqtSignal(bdk.Transaction)

    def __init__(
        self,
        signature_importers: Iterable[AbstractSignatureImporter],
        psbt: bdk.PartiallySignedTransaction,
        network: bdk.Network,
    ) -> None:
        super().__init__()
        self.signature_importers = signature_importers
        self.psbt = psbt
        self.network = network

        self.layout_keystore_buttons = QVBoxLayout(self)

        def callback_generator(signer: AbstractSignatureImporter) -> Callable:
            def f() -> None:
                signer.sign(self.psbt)

            return f

        for signer in self.signature_importers:
            button = QPushButton(signer.label)
            button.setMinimumHeight(30)
            button.setIcon(QIcon(icon_path(signer.keystore_type.icon_filename)))
            button.clicked.connect(callback_generator(signer))
            self.layout_keystore_buttons.addWidget(button)

            # forward the signal_signature_added from each signer to self.signal_signature_added
            signer.signal_signature_added.connect(self.signal_signature_added.emit)
            signer.signal_final_tx_received.connect(self.signal_tx_received.emit)


class SignerUIHorizontal(QWidget):
    signal_signature_added = pyqtSignal(bdk.PartiallySignedTransaction)
    signal_tx_received = pyqtSignal(bdk.Transaction)

    def __init__(
        self,
        signature_importers: List[AbstractSignatureImporter],
        psbt: bdk.PartiallySignedTransaction,
        network: bdk.Network,
    ) -> None:
        super().__init__()
        self.signature_importers = signature_importers
        self.psbt = psbt
        self.network = network

        self.layout_keystore_buttons = QVBoxLayout(self)

        for signer in self.signature_importers:

            def callback_generator(signer: AbstractSignatureImporter) -> Callable:
                def f() -> None:
                    signer.sign(self.psbt)

                return f

            button = QPushButton(signer.label)
            button.setMinimumHeight(30)
            button.setIcon(QIcon(icon_path(signer.keystore_type.icon_filename)))
            button.clicked.connect(callback_generator(signer))
            self.layout_keystore_buttons.addWidget(button)

            # forward the signal_signature_added from each signer to self.signal_signature_added
            signer.signal_signature_added.connect(self.signal_signature_added.emit)
            signer.signal_final_tx_received.connect(self.signal_tx_received.emit)



#
# Bitcoin Safe
# Copyright (C) 2024 Andreas Griffin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/gpl-3.0.html
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import logging

from ....i18n import translate
from ....util import qbytearray_to_str, register_cache, str_to_qbytearray

logger = logging.getLogger(__name__)

import hashlib
import json
from typing import Dict, Generator, Iterable, List, Optional, Tuple

from PyQt6.QtCore import QMimeData, QModelIndex, QPoint, QRect, QSize, Qt, pyqtSignal
from PyQt6.QtGui import (
    QColor,
    QCursor,
    QDrag,
    QDragEnterEvent,
    QDragLeaveEvent,
    QDropEvent,
    QFont,
    QImage,
    QMouseEvent,
    QPainter,
    QPalette,
    QTextDocument,
    QTextOption,
)
from PyQt6.QtWidgets import (
    QAbstractItemView,
    QApplication,
    QLineEdit,
    QListWidget,
    QListWidgetItem,
    QPushButton,
    QStyle,
    QStyledItemDelegate,
    QStyleOptionButton,
    QStyleOptionViewItem,
    QVBoxLayout,
    QWidget,
)


def clean_tag(tag: str) -> str:
    return tag.strip()


class AddressDragInfo:
    def __init__(self, tags: Iterable[Optional[str]], addresses: List[str]) -> None:
        self.tags = tags
        self.addresses = addresses

    def __repr__(self) -> str:
        return f"AddressDragInfo({self.tags}, {self.addresses})"


def hash_string(text: str) -> str:
    return hashlib.sha256(str(text).encode()).hexdigest()


def rescale(value: float, old_min: float, old_max: float, new_min: float, new_max: float):
    return (value - old_min) / (old_max - old_min) * (new_max - new_min) + new_min


@register_cache(always_keep=True)
def hash_color(text) -> QColor:
    hash_value = int(hash_string(text), 16) & 0xFFFFFF
    r = (hash_value & 0xFF0000) >> 16
    g = (hash_value & 0x00FF00) >> 8
    b = hash_value & 0x0000FF

    r = int(rescale(r, 0, 255, 100, 255))
    g = int(rescale(g, 0, 255, 100, 255))
    b = int(rescale(b, 0, 255, 100, 255))

    return QColor(r, g, b)


class CustomListWidgetItem(QListWidgetItem):
    def __init__(self, item_text: str, sub_text: str | None = None, parent=None):
        super(CustomListWidgetItem, self).__init__(parent)
        self.setText(item_text)
        self.subtext = sub_text
        self.color = self.hash_color()
        self.setData(Qt.ItemDataRole.UserRole + 1, self.color)
        self.setData(Qt.ItemDataRole.UserRole + 2, self.subtext)  # UserRole for subtext

    def hash_color(self):
        return hash_color(self.text())

    def mimeData(self):
        mime_data = QMimeData()
        d = {
            "type": "drag_tag",
            "tag": self.text(),
        }

        mime_data.setData("application/json", str_to_qbytearray(json.dumps(d)))
        return mime_data


class CustomDelegate(QStyledItemDelegate):
    signal_tag_renamed = pyqtSignal(str, str)

    def __init__(self, parent) -> None:
        super().__init__(parent)
        self.currentlyEditingIndex = QModelIndex()
        self.imageCache: Dict[Tuple[QModelIndex, QStyle.StateFlag, str, str, int, int], QImage] = (
            {}
        )  # Cache for storing pre-rendered images
        self.cache_size = 100

    def _remove_first_cache_item(self):
        key = None
        for key in self.imageCache.keys():
            break
        if key in self.imageCache:
            del self.imageCache[key]

    def add_to_cache(self, key, value):
        # logger.debug(f"Cache size {self.__class__.__name__}: {len(self.imageCache)}")
        if len(self.imageCache) + 1 > self.cache_size:
            self._remove_first_cache_item()

        self.imageCache[key] = value

    def renderHtmlToImage(self, index: QModelIndex, option: QStyleOptionViewItem, text: str, subtext: str):
        """
        Renders the item appearance to an image, including button-like background and HTML text.
        """
        rectSize = QSize(option.rect.width(), option.rect.height())

        # Create an off-screen image for rendering
        image = QImage(rectSize, QImage.Format.Format_ARGB32_Premultiplied)
        image.fill(Qt.GlobalColor.transparent)  # Fill with transparency for proper rendering
        painter = QPainter(image)

        # Mock-up style option for button rendering
        buttion_option = QStyleOptionButton()
        buttion_option.rect = QRect(0, 0, rectSize.width(), rectSize.height())
        buttion_option.state = QStyle.StateFlag.State_Enabled

        if option.state & QStyle.StateFlag.State_Selected:
            buttion_option.state |= QStyle.StateFlag.State_Sunken
        else:
            buttion_option.state |= QStyle.StateFlag.State_Raised

        color = QColor(index.data(Qt.ItemDataRole.UserRole + 1))  # Assuming color is stored in UserRole + 1
        buttion_option.palette.setColor(QPalette.ColorRole.Button, color)

        # Draw button-like background
        (QApplication.style() or QStyle()).drawControl(
            QStyle.ControlElement.CE_PushButton, buttion_option, painter
        )

        # Render HTML text
        self.draw_html_text(painter, text, subtext, buttion_option.rect, scale=1)

        painter.end()
        return image

    def draw_html_text(self, painter: QPainter, text: str, subtext: str, rect: QRect, scale: float):
        # Base font size adjustments (these are scale factors, adjust as needed)
        baseFontSize = 1.0  # Adjust this factor to scale the default font size for main text
        subtextFontSize = 0.8  # Adjust this factor for subtext

        # Get the default system font
        defaultFont = QApplication.font()

        # Calculate text and subtext positions more accurately
        textPadding = 4  # Adjust padding as needed
        subtextPadding = 2  # Adjust subtext padding

        # Height allocation (tweak these ratios as needed)
        textHeightRatio = 0.6  # Allocate 60% of the rect height to the main text
        subtextHeightRatio = 0.4  # Remaining 40% for the subtext

        # Main text area
        rectText = QRect(
            rect.left() + textPadding,
            rect.top() + textPadding,
            rect.width() - 2 * textPadding,
            int(rect.height() * textHeightRatio) - textPadding,
        )

        # Subtext area
        rectSubtext = QRect(
            rect.left() + subtextPadding,
            rect.top() + int(rect.height() * textHeightRatio),
            rect.width() - 2 * subtextPadding,
            int(rect.height() * subtextHeightRatio) - subtextPadding,
        )

        # Draw main text
        if text:
            doc = QTextDocument()
            doc.setHtml(text)
            font = QFont(defaultFont)  # Use the default system font
            font.setPointSizeF(defaultFont.pointSizeF() * baseFontSize)  # Apply scaling factor
            doc.setDefaultFont(font)
            textOption = QTextOption()
            textOption.setAlignment(Qt.AlignmentFlag.AlignCenter)  # Center text horizontally
            doc.setDefaultTextOption(textOption)
            painter.save()
            painter.translate(rectText.topLeft())
            doc.setTextWidth(rectText.width())
            # Optional: Adjust for vertical centering
            yOffset = (rectText.height() - doc.size().height()) / 2
            painter.translate(0, max(yOffset, 0))
            doc.drawContents(painter)
            painter.restore()

        # Draw subtext
        if subtext:
            doc = QTextDocument()
            doc.setHtml(subtext)
            font = QFont(defaultFont)  # Use the default system font for subtext
            font.setPointSizeF(defaultFont.pointSizeF() * subtextFontSize)  # Apply scaling factor for subtext
            doc.setDefaultFont(font)
            textOption = QTextOption()
            textOption.setAlignment(Qt.AlignmentFlag.AlignCenter)  # Center text horizontally
            doc.setDefaultTextOption(textOption)
            painter.save()
            painter.translate(rectSubtext.topLeft())
            doc.setTextWidth(rectSubtext.width())
            # Optional: Adjust for vertical centering
            yOffset = (rectSubtext.height() - doc.size().height()) / 2
            painter.translate(0, max(yOffset, 0))
            painter.scale(scale, scale)  # Apply scaling if needed
            doc.drawContents(painter)
            painter.restore()

    def paint(self, painter: QPainter | None, option: QStyleOptionViewItem, index: QModelIndex):
        if not painter:
            super().paint(painter=painter, option=option, index=index)
            return

        # Check if the editor is open for this index
        if self.currentlyEditingIndex.isValid() and self.currentlyEditingIndex == index:
            text = ""  # Set text to empty if editor is open
            subtext = ""
        else:
            text = index.data(Qt.ItemDataRole.DisplayRole)
            subtext = index.data(Qt.ItemDataRole.UserRole + 2)  # Assuming subtext is stored in UserRole + 2

        key = (
            index,
            option.state,
            text,
            subtext,
            option.rect.width(),
            option.rect.height(),
        )
        # Ensure there's an image rendered for this index
        if key not in self.imageCache:
            # Render and cache the item appearance
            self.add_to_cache(key, self.renderHtmlToImage(index, option, text, subtext))

        # Draw the cached image
        image = self.imageCache[key]
        if image:
            painter.drawImage(option.rect.topLeft(), image)

    def clearCache(self):
        """
        Clears the cached images. Call this method if you need to refresh the items.
        """
        self.imageCache.clear()

    def createEditor(self, parent, option: QStyleOptionViewItem, index: QModelIndex):
        self.currentlyEditingIndex = index
        editor = QLineEdit(parent)
        editor.setAlignment(Qt.AlignmentFlag.AlignCenter)
        editor.setStyleSheet(
            """
            background: transparent;
            border: none;
        """
        )
        return editor

    def setEditorData(self, editor: QLineEdit, index: QModelIndex):  # type: ignore[override]
        model = index.model()
        if not model:
            return
        value = model.data(index, Qt.ItemDataRole.EditRole)
        editor.setText(value)

    def setModelData(self, editor: QLineEdit, model, index: QModelIndex):  # type: ignore[override]
        model = index.model()
        if not model:
            return
        old_value = model.data(index, Qt.ItemDataRole.EditRole)
        new_value = clean_tag(editor.text())

        model.setData(index, editor.text(), Qt.ItemDataRole.EditRole)
        self.currentlyEditingIndex = QModelIndex()
        self.signal_tag_renamed.emit(old_value, new_value)


class DeleteButton(QPushButton):
    signal_delete_item = pyqtSignal(str)
    signal_addresses_dropped = pyqtSignal(AddressDragInfo)

    def __init__(self, *args, **kwargs):
        super(DeleteButton, self).__init__(*args, **kwargs)
        self.setAcceptDrops(True)
        icon = (self.style() or QStyle()).standardIcon(QStyle.StandardPixmap.SP_TrashIcon)
        self.setIcon(icon)

    def dragEnterEvent(self, event: QDragEnterEvent | None):
        if not event:
            super().dragEnterEvent(event)
            return

        mime_data = event.mimeData()
        if mime_data and mime_data.hasFormat("application/json"):
            json_string = qbytearray_to_str(mime_data.data("application/json"))

            d = json.loads(json_string)
            logger.debug(f"dragEnterEvent: Got {d}")
            if d.get("type") == "drag_tag" or d.get("type") == "drag_addresses":
                event.acceptProposedAction()
                return

        event.ignore()

    def dragLeaveEvent(self, event: QDragLeaveEvent | None):
        "this is just to hide/undide the button"
        logger.debug("Drag has left the delete button")

    def dropEvent(self, event: QDropEvent | None):
        super().dropEvent(event)
        if not event or event.isAccepted():
            return

        mime_data = event.mimeData()
        if mime_data and mime_data.hasFormat("application/json"):
            json_string = qbytearray_to_str(mime_data.data("application/json"))

            d = json.loads(json_string)
            logger.debug(f"dropEvent: Got {d}")
            if d.get("type") == "drag_tag":
                self.signal_delete_item.emit(d.get("tag"))
                event.acceptProposedAction()
                return
            if d.get("type") == "drag_addresses":
                drag_info = AddressDragInfo([None], d.get("addresses"))
                logger.debug(f"dropEvent: {drag_info}")
                self.signal_addresses_dropped.emit(drag_info)
                event.accept()
                return

        event.ignore()


class CustomListWidget(QListWidget):
    signal_tag_added = pyqtSignal(str)
    signal_tag_clicked = pyqtSignal(str)
    signal_tag_deleted = pyqtSignal(str)
    signal_tag_renamed = pyqtSignal(str, str)
    signal_addresses_dropped = pyqtSignal(AddressDragInfo)
    signal_start_drag = pyqtSignal(object)
    signal_stop_drag = pyqtSignal(object)

    def __init__(self, parent=None, enable_drag=True, immediate_release=True):
        super().__init__(parent)

        self.immediate_release = immediate_release

        delegate = CustomDelegate(self)
        delegate.signal_tag_renamed.connect(self.signal_tag_renamed)
        self.setItemDelegate(delegate)
        self.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)
        self.setSelectionBehavior(QListWidget.SelectionBehavior.SelectItems)

        self.setAcceptDrops(True)
        if viewport := self.viewport():
            viewport.setAcceptDrops(True)
        self.setDropIndicatorShown(True)
        self.setDragDropMode(QAbstractItemView.DragDropMode.InternalMove)
        self.setDefaultDropAction(Qt.DropAction.CopyAction)
        self.setDragEnabled(enable_drag)  # this must be after the other drag toggles

        self.itemChanged.connect(self.on_item_changed)  # new

        self.setMouseTracking(True)
        self._drag_start_position = None

        self.setStyleSheet(
            """
            QListWidget {
                background: transparent;
                border: none;
            }
            QListWidget::item {
                border-radius: 5px;
                margin: 15px;
            }
            QListWidget::item:selected {
                background-color: qlineargradient(x1:0, y1:0, x2:0, y2:1, stop:0 #cccccc, stop:1 #b3b3b3);
                border: 1px solid black;
            }
            """
        )

    def add(self, item_text: str, sub_text=None) -> CustomListWidgetItem:
        item = CustomListWidgetItem(item_text, sub_text=sub_text)
        item.setFlags(item.flags() | Qt.ItemFlag.ItemIsEditable)
        self.addItem(item)
        self.signal_tag_added.emit(item_text)
        return item

    def on_item_clicked(self, item: QListWidgetItem):
        self.signal_tag_clicked.emit(item.text())
        # print( [item.text() for item in self.selectedItems()])

    def on_item_changed(self, item: CustomListWidgetItem):  # new

        item.color = item.hash_color()
        item.setData(Qt.ItemDataRole.UserRole + 1, item.color)

        # Here you can handle the renaming event
        # For now, we will just print the new item text
        # print(f"Item text has been changed to {item.text()}")

    def get_selected(self) -> List[str]:
        return [item.text() for item in self.selectedItems()]

    def rename_selected(self, new_text: str):
        for item in self.selectedItems():
            item.text()
            item.setText(new_text)
            # item.setBackground()

    def setAllSelection(self, selected=True):
        for i in range(self.count()):
            item = self.item(i)
            if item:
                item.setSelected(selected)

    def mousePressEvent(self, event: QMouseEvent | None):
        if not event:
            super().mousePressEvent(event)
            return

        item = self.itemAt(event.pos())
        if item and event.button() == Qt.MouseButton.LeftButton:
            self._drag_start_position = event.pos()
            if not (QApplication.keyboardModifiers() & Qt.KeyboardModifier.ControlModifier):
                if item.isSelected():
                    self.setAllSelection(False)
                    return

        super().mousePressEvent(event)

    def mouseDoubleClickEvent(self, event: QMouseEvent | None):
        if event and event.button() == Qt.MouseButton.LeftButton:
            pass

        super().mouseDoubleClickEvent(event)

    def mouseReleaseEvent(self, event: QMouseEvent | None):
        if event and event.button() == Qt.MouseButton.LeftButton:
            # Perform actions that should happen after the mouse button is released
            # This could be updating the state of the widget, triggering signals, etc.

            item = self.itemAt(event.pos())
            if item is not None and item.isSelected():
                self.on_item_clicked(item)
                if self.immediate_release:
                    if not (QApplication.keyboardModifiers() & Qt.KeyboardModifier.ControlModifier):
                        self.setAllSelection(False)
                        return

        super().mouseReleaseEvent(event)

    def mouseMoveEvent(self, event: QMouseEvent | None):
        if not event:
            super().mouseMoveEvent(event)
            return

        if self._drag_start_position is None:
            self._drag_start_position = event.pos()
        if not (event.buttons() & Qt.MouseButton.LeftButton):
            return
        if (event.pos() - self._drag_start_position).manhattanLength() < QApplication.startDragDistance():
            return
        if self.dragEnabled():
            self.startDrag(Qt.DropAction.MoveAction)

        super().mouseMoveEvent(event)

    def startDrag(self, action: Qt.DropAction | None):
        item = self.currentItem()
        if not action or not isinstance(item, CustomListWidgetItem):
            return
        rect = self.visualItemRect(item)

        drag = QDrag(self)
        drag.setMimeData(item.mimeData())

        viewport = self.viewport()
        if not viewport:
            return
        pixmap = viewport.grab(rect)
        hotspot_pos = QPoint(0, 0)
        drag.setPixmap(pixmap)
        drag.setHotSpot(hotspot_pos)

        self.signal_start_drag.emit(action)

        drag.exec(action)

        self.signal_stop_drag.emit(action)

    def dragEnterEvent(self, event: QDragEnterEvent | None):
        if not event:
            super().dragEnterEvent(event)
            return

        mime_data = event.mimeData()
        if mime_data and mime_data.hasFormat("application/json"):
            # print('accept')
            # tag = self.itemAt(event.pos())

            json_string = qbytearray_to_str(mime_data.data("application/json"))
            # dropped_addresses = json.loads(json_string)
            # print(f'drag enter {dropped_addresses,   tag.text()}')
            logger.debug(f"dragEnterEvent: {json_string}")

            event.acceptProposedAction()
        else:
            event.ignore()

    def dropEvent(self, event: QDropEvent | None):
        logger.debug("drop")
        super().dropEvent(event)
        if not event or event.isAccepted():
            return

        mime_data = event.mimeData()
        if mime_data and mime_data.hasFormat("application/json"):
            tag = self.itemAt(event.position().toPoint())

            json_string = qbytearray_to_str(mime_data.data("application/json"))

            d = json.loads(json_string)
            if d.get("type") == "drag_addresses":
                if tag is not None:
                    drag_info = AddressDragInfo([tag.text()], d.get("addresses"))
                    logger.debug(f"dropEvent: {drag_info}")
                    self.signal_addresses_dropped.emit(drag_info)
                event.accept()
                return

        event.ignore()

    def delete_item(self, item_text: str):
        for i in range(self.count()):
            item = self.item(i)
            if item and item.text() == item_text:
                self.takeItem(i)
                self.signal_tag_deleted.emit(item_text)
                break

    def get_items(self) -> Generator[QListWidgetItem, None, None]:
        for i in range(self.count()):
            item = self.item(i)
            if item:
                yield item

    def get_item_texts(self) -> Generator[str, None, None]:
        for item in self.get_items():
            yield item.text()

    def recreate(self, tags: List[str], sub_texts: Iterable[str | None] | None = None):
        # Store the texts of selected items
        selected_texts = [item.text() for item in self.selectedItems()]

        # Delete all items
        for i in reversed(range(self.count())):
            self.takeItem(i)

        # Add all items back
        cleaned_sub_texts = sub_texts if sub_texts else [None] * len(tags)
        for sub_text, tag in zip(cleaned_sub_texts, tags):
            self.add(tag, sub_text=sub_text)  # Assuming `self.add` correctly adds items

        # Re-select items based on stored texts
        for i in range(self.count()):
            item = self.item(i)
            if item and item.text() in selected_texts:
                item.setSelected(True)


class TagEditor(QWidget):
    def __init__(
        self,
        parent=None,
        tags: List[str] | None = None,
        sub_texts: Iterable[str | None] | None = None,
        tag_name="tag",
    ):
        super(TagEditor, self).__init__(parent)
        self.tag_name = tag_name
        self._layout = QVBoxLayout(self)

        self.input_field = QLineEdit()
        self.input_field.setClearButtonEnabled(True)
        self.input_field.returnPressed.connect(self.add_new_tag_from_input_field)

        self.delete_button = DeleteButton()
        self.delete_button.hide()

        self.list_widget = CustomListWidget(parent=self)
        self._layout.addWidget(self.input_field)
        self._layout.addWidget(self.delete_button)
        self._layout.addWidget(self.list_widget)
        self.list_widget.signal_start_drag.connect(self.show_delete_button)
        self.list_widget.signal_stop_drag.connect(self.hide_delete_button)
        self.list_widget.signal_addresses_dropped.connect(self.hide_delete_button)
        self.delete_button.signal_delete_item.connect(self.list_widget.delete_item)
        self.delete_button.signal_addresses_dropped.connect(self.hide_delete_button)

        self.setAcceptDrops(True)
        # TagEditor.updateUi  ensure that this is not overwritten in a child class
        TagEditor.updateUi(self)

        if tags:
            self.list_widget.recreate(tags, sub_texts=sub_texts)

    def updateUi(self):
        self.input_field.setPlaceholderText(self.default_placeholder_text())
        self.delete_button.setText(translate("tageditor", "Delete {name}").format(name=self.tag_name))

    def default_placeholder_text(self):
        return translate("tageditor", "Add new {name}").format(name=self.tag_name)

    def dragEnterEvent(self, event: QDragEnterEvent | None):
        if not event:
            super().dragEnterEvent(event)
            return

        mime_data = event.mimeData()
        if mime_data and mime_data.hasFormat("application/json"):
            # print('accept')
            # tag = self.itemAt(event.pos())

            json_string = qbytearray_to_str(mime_data.data("application/json"))
            # dropped_addresses = json.loads(json_string)
            # print(f'drag enter {dropped_addresses,   tag.text()}')
            logger.debug(f"dragEnterEvent: {json_string}")

            event.acceptProposedAction()

            logger.debug(f"show_delete_button")
            self.show_delete_button()
        else:
            event.ignore()

    def dragLeaveEvent(self, event: QDragLeaveEvent | None):
        "this is just to hide/undide the button"
        if not event:
            super().dragLeaveEvent(event)
            return

        if not self.rect().contains(self.mapFromGlobal(QCursor.pos())):
            logger.debug("Drag operation left the TagEditor")
            self.hide_delete_button()
        else:
            event.ignore()

    def dropEvent(self, event: QDropEvent | None):
        super().dropEvent(event)

        if not event or event.isAccepted():
            return

        mime_data = event.mimeData()
        if mime_data and mime_data.hasFormat("application/json"):
            self.hide_delete_button()
            event.accept()
        else:
            event.ignore()

    def show_delete_button(self, *args):
        self.input_field.hide()
        self.delete_button.show()

    def hide_delete_button(self, *args):
        self.input_field.show()
        self.delete_button.hide()

    def add(self, new_tag: str, sub_text: str | None = None) -> Optional[CustomListWidgetItem]:
        if not self.tag_exists(new_tag):
            return self.list_widget.add(new_tag, sub_text=sub_text)
        return None

    def add_new_tag_from_input_field(self):
        new_tag = clean_tag(self.input_field.text())
        item = self.add(new_tag)
        if item:
            self.input_field.setPlaceholderText(self.default_placeholder_text())
        else:
            self.input_field.setPlaceholderText(
                translate("tageditor", "This {name} exists already.").format(name=self.tag_name)
            )
        self.input_field.clear()

    def tag_exists(self, tag: str):
        for i in range(self.list_widget.count()):
            item = self.list_widget.item(i)
            if item and item.text() == tag:
                return True
        return False



#
# Bitcoin Safe
# Copyright (C) 2024 Andreas Griffin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/gpl-3.0.html
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import logging

from bitcoin_safe.html import html_f
from bitcoin_safe.i18n import translate
from bitcoin_safe.signals import SignalsMin

logger = logging.getLogger(__name__)

import sys
from typing import Callable, List, Optional

from PyQt6.QtCore import QEvent, QModelIndex, QObject, QPoint, Qt
from PyQt6.QtGui import (
    QKeyEvent,
    QPainter,
    QStandardItem,
    QStandardItemModel,
    QTextDocument,
)
from PyQt6.QtWidgets import (
    QAbstractItemView,
    QApplication,
    QFrame,
    QLineEdit,
    QMainWindow,
    QPushButton,
    QStyle,
    QStyledItemDelegate,
    QStyleOptionViewItem,
    QTabWidget,
    QTreeView,
    QVBoxLayout,
    QWidget,
)

from bitcoin_safe.gui.qt.my_treeview import MyTreeView, SearchableTab
from bitcoin_safe.gui.qt.qt_wallet import QTWallet
from bitcoin_safe.gui.qt.ui_tx import UITx_Creator


class SearchHTMLDelegate(QStyledItemDelegate):
    def __init__(self, parent: QWidget) -> None:
        super().__init__(parent)
        self.setParent(parent=parent)

    def setParent(self, parent: "QWidget") -> None:  # type: ignore[override]
        self._parent = parent
        super().setParent(parent)

    def parent(self) -> QWidget:
        return self._parent

    def paint(self, painter: QPainter, option: QStyleOptionViewItem, index: QModelIndex) -> None:  # type: ignore[override]
        model = index.model()
        if not model:
            return
        text = model.data(index, Qt.ItemDataRole.DisplayRole)

        # Use QStyle to draw the item. This respects the native theme.
        (self.parent().style() or QStyle()).drawPrimitive(
            QStyle.PrimitiveElement.PE_PanelItemViewItem, option, painter, self.parent()
        )

        painter.save()

        # Set up the QTextDocument for HTML rendering
        document = QTextDocument()
        document.setHtml(text)

        # Calculate the vertical alignment to center the text
        text_height = document.size().height()
        center_y = option.rect.center().y() - (text_height / 2)

        painter.translate(option.rect.left(), center_y)
        document.drawContents(painter)

        painter.restore()


class ResultItem:
    def __init__(self, text: str, parent: Optional["ResultItem"] = None, obj=None, obj_key=None) -> None:
        self.text = text
        self.obj = obj
        self.obj_key = obj_key
        self.children: List["ResultItem"] = []

        self.set_parent(parent)

    def set_parent(self, parent: Optional["ResultItem"] = None) -> None:
        self.parent = parent
        if self.parent:
            if self not in self.parent.children:
                self.parent.children.append(self)


def demo_do_search(search_text: str) -> ResultItem:
    # Demo search function. Replace with actual search logic
    # Returns data in the format expected by CustomTreeView.set_data
    root = ResultItem("")

    if not search_text:
        return root

    search_text = search_text.strip()

    wallet = ResultItem("test", parent=root)
    addresses = ResultItem("addresses", parent=wallet)
    utxo = ResultItem("utxo", parent=wallet)
    history = ResultItem("history", parent=wallet)

    for l in [addresses, utxo, history]:
        for txt in ["aaaa", "bbbb"]:
            text = txt + search_text + txt
            ResultItem(text, parent=l)

    return root


def demo_on_click(item: ResultItem) -> None:
    print("Item Clicked:", item.text)


class CustomItem(QStandardItem):
    def __init__(self, *args) -> None:
        super().__init__(*args)
        self.result_item: Optional[ResultItem] = None


class CustomItemModel(QStandardItemModel):
    def invisibleRootItem(self) -> CustomItem | None:
        return super().invisibleRootItem()  # type: ignore


class CustomTreeView(QTreeView):
    def __init__(self, parent=None, on_click=None, on_double_click=None) -> None:
        super().__init__(parent)
        self.on_click = on_click
        self.on_double_click = on_double_click
        self.setModel(CustomItemModel())

        self.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAsNeeded)  # Vertical scrollbar
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAsNeeded)  # Horizontal scrollbar

        self.setHeaderHidden(True)

        # Set the selection behavior to select full rows

        self.setSelectionMode(
            QAbstractItemView.SelectionMode.SingleSelection
        )  # Allow selecting, but not editing
        self.setEditTriggers(QAbstractItemView.EditTrigger.NoEditTriggers)  # Disable editing

        # Make the last column stretch to fill the view
        # self.header().setStretchLastSection(True)
        # self.header().setSectionResizeMode(QHeaderView.ResizeToContents)

        # Connect the clicked signal to the callback function
        self.clicked.connect(self.handle_item_clicked)
        self.doubleClicked.connect(self.handle_item_double_clicked)

    def setModel(self, model: Optional[CustomItemModel]) -> None:  # type: ignore[override]
        self._source_model = model
        super().setModel(model)

    def model(self) -> CustomItemModel:
        if self._source_model:
            return self._source_model
        raise Exception("model not set")

    def set_data(self, data: ResultItem) -> None:
        self.model().clear()  # Clear existing items
        self._populate_model(data)
        self.expandAll()  # Expand all items after setting data
        self.resizeColumnToContents(0)  # Resize the first column

    def _populate_model(self, result_item: ResultItem, model_parent: CustomItem | None = None) -> None:
        def add_child(child: ResultItem) -> CustomItem:
            model_item = CustomItem(child.text)
            model_item.setEditable(False)
            model_item.result_item = child
            if model_parent:
                model_parent.appendRow(model_item)
            return model_item

        model_parent = (
            self.model().invisibleRootItem() if model_parent is None else add_child(result_item)
        )  # type: ignore[assignment]

        for child in result_item.children:
            # Recursively process the value
            self._populate_model(child, model_parent=model_parent)

    def handle_item_clicked(self, index: QModelIndex) -> None:
        if self.on_click and index.isValid():
            # Retrieve the item from the model
            item = self.model().itemFromIndex(index)
            if not item:
                return
            if not isinstance(item, CustomItem):
                logger.error(f"{item} has wrong type  {type(item)} != CustomItem")
                return
            # Perform the action you want based on the clicked item
            # For example, call a custom method of the item (if your item class has one)
            self.on_click(item.result_item)

    def handle_item_double_clicked(self, index: QModelIndex) -> None:
        if self.on_double_click and index.isValid():
            # Retrieve the item from the model
            item = self.model().itemFromIndex(index)
            if not item:
                return
            if not isinstance(item, CustomItem):
                logger.error(f"{item} has wrong type  {type(item)} != CustomItem")
                return
            # Perform the action you want based on the clicked item
            # For example, call a custom method of the item (if your item class has one)
            self.on_double_click(item.result_item)


class CustomPopup(QFrame):
    def __init__(self, parent=None) -> None:
        super(CustomPopup, self).__init__(parent, Qt.WindowType.Tool | Qt.WindowType.FramelessWindowHint)
        self._layout = QVBoxLayout(self)
        self._layout.setContentsMargins(0, 1, 0, 0)  # Left, Top, Right, Bottom margins

        # self.setFrameShape(QFrame.Panel)

        self.hide()  # Start hidden

    # Override keyPressEvent method
    def keyPressEvent(self, event: QKeyEvent) -> None:  # type: ignore[override]
        # Check if the pressed key is 'Esc'
        if event.key() == Qt.Key.Key_Escape:
            # Close the widget
            self.hide()


class SearchTreeView(QWidget):
    prev_main_window_pos = None

    def __init__(
        self,
        do_search: Callable[[str], ResultItem],
        parent=None,
        results_in_popup=True,
        on_click: Optional[Callable] = None,
        result_width=None,
        result_height=300,
    ) -> None:
        super().__init__(parent)
        self.on_click = on_click
        self.do_search = do_search
        self.result_width = result_width
        self.result_height = result_height
        self._layout = QVBoxLayout(self)
        self._layout.setContentsMargins(0, 0, 0, 0)

        self.search_field = QLineEdit(self)
        self.search_field.setClearButtonEnabled(True)
        self._layout.addWidget(self.search_field)

        self.popup = CustomPopup(self)

        self.tree_view = CustomTreeView(self, on_click=on_click, on_double_click=self.on_double_click)
        self.tree_view.setVisible(False)
        if results_in_popup:
            self.popup._layout.addWidget(self.tree_view)
        else:
            self._layout.addWidget(self.tree_view)

        self.search_field.textChanged.connect(self.on_search)

        self.highlight_delegate = SearchHTMLDelegate(self.tree_view)
        self.tree_view.setItemDelegate(self.highlight_delegate)

        self.updateUi()
        # Install event filter on the main window
        window = self.window()
        if window:
            window.installEventFilter(self)

    def on_double_click(self, result_item: ResultItem) -> None:
        "Here is what is done on the 2. click of a double click"
        self.popup.hide()

    def updateUi(self) -> None:
        self.search_field.setPlaceholderText(translate("search_treeview", "Type to search..."))

    def on_search(self, text: str) -> None:
        search_results = self.do_search(text)
        self.tree_view.set_data(search_results)
        self.tree_view.update()  # Update the view to redraw with highlights
        self.tree_view.setVisible(bool(search_results))

        if text:
            self.position_popup()
            self.popup.show()
        else:
            self.popup.hide()

    def position_popup(self) -> None:
        self.popup.setFixedSize(
            self.search_field.width() if self.result_width is None else self.result_width, self.result_height
        )  # Set a fixed size or adjust as needed

        # Calculate the global position for the popup
        global_pos = self.search_field.mapToGlobal(
            QPoint(int(self.search_field.width() - self.popup.width()), int(self.search_field.height()))
        )
        self.popup.move(global_pos)

    # Override keyPressEvent method
    def keyPressEvent(self, event: QKeyEvent) -> None:  # type: ignore[override]
        # Check if the pressed key is 'Esc'
        if event.key() == Qt.Key.Key_Escape:
            # Close the widget
            self.popup.hide()

    def eventFilter(self, obj: QObject, event: QEvent) -> bool:  # type: ignore[override]
        if obj in [self, self.window()]:
            if event.type() == QEvent.Type.Move or event.type() == QEvent.Type.Resize:
                self.position_popup()
        return super().eventFilter(obj, event)


class SearchWallets(SearchTreeView):
    def __init__(
        self,
        get_qt_wallets: Callable[[], List[QTWallet]],
        signal_min: SignalsMin,
        parent=None,
        result_height=300,
        result_width=500,
    ) -> None:
        super().__init__(
            self.do_search,
            parent=parent,
            results_in_popup=True,
            on_click=self.search_result_on_click,
            result_height=result_height,
            result_width=result_width,
        )
        self.signal_min = signal_min

        self.get_qt_wallets = get_qt_wallets
        self.signal_min.language_switch.connect(self.updateUi)

    def search_result_on_click(self, result_item: ResultItem) -> None:
        # call the parent action first
        if result_item.parent is not None:
            self.search_result_on_click(result_item.parent)

        if isinstance(result_item.obj, MyTreeView):
            result_item.obj.select_row(result_item.obj_key, result_item.obj.key_column)
        elif isinstance(result_item.obj, SearchableTab):
            parent = result_item.obj.parent()
            if parent:
                tabs = parent.parent()
                if isinstance(tabs, QTabWidget):
                    tabs.setCurrentWidget(result_item.obj)
        elif isinstance(result_item.obj, UITx_Creator):
            parent = result_item.obj.parent()
            if parent:
                these_tabs = parent.parent()
                if isinstance(these_tabs, QTabWidget):
                    these_tabs.setCurrentWidget(result_item.obj)
            result_item.obj.tabs_inputs.setCurrentWidget(result_item.obj.tab_inputs_utxos)
        elif isinstance(result_item.obj, QTWallet):
            parent = result_item.obj.tab.parent()
            if parent:
                wallet_tabs = parent.parent()
                if isinstance(wallet_tabs, QTabWidget):
                    wallet_tabs.setCurrentWidget(result_item.obj.tab)

    def do_search(self, search_text: str) -> ResultItem:
        def format_result_text(matching_string: str) -> str:
            return matching_string.replace(
                search_text, f"<span style='background-color: #ADD8E6;'>{search_text}</span>"
            )

        search_text = search_text.strip()
        root = ResultItem("")
        for qt_wallet in self.get_qt_wallets():
            wallet_item = ResultItem(html_f(qt_wallet.wallet.id, bf=True), obj=qt_wallet)

            wallet_addresses = ResultItem(html_f(self.tr("Addresses"), bf=True), obj=qt_wallet.addresses_tab)
            for address in qt_wallet.wallet.get_addresses():
                if search_text in address:
                    ResultItem(
                        format_result_text(address),
                        parent=wallet_addresses,
                        obj=qt_wallet.address_list,
                        obj_key=address,
                    )
                    # connect also the higher parents, so the results appear at all
                    wallet_addresses.set_parent(wallet_item)
                    wallet_item.set_parent(root)
                label = qt_wallet.wallet.get_label_for_address(address, autofill_from_txs=False)
                if label and (search_text.lower() in label.lower()):
                    ResultItem(
                        format_result_text(label),
                        parent=wallet_addresses,
                        obj=qt_wallet.address_list,
                        obj_key=address,
                    )
                    # connect also the higher parents, so the results appear at all
                    wallet_addresses.set_parent(wallet_item)
                    wallet_item.set_parent(root)

            wallet_tx_ids = ResultItem(html_f(self.tr("Transactions"), bf=True), obj=qt_wallet.history_tab)
            for txid, fulltxdetail in qt_wallet.wallet.get_dict_fulltxdetail().items():
                if search_text in txid:
                    ResultItem(
                        format_result_text(txid),
                        parent=wallet_tx_ids,
                        obj=qt_wallet.history_list,
                        obj_key=txid,
                    )
                    # connect also the higher parents, so the results appear at all
                    wallet_tx_ids.set_parent(wallet_item)
                    wallet_item.set_parent(root)
                label = qt_wallet.wallet.get_label_for_txid(txid, autofill_from_addresses=False)
                if label and (search_text.lower() in label.lower()):
                    ResultItem(
                        format_result_text(label),
                        parent=wallet_tx_ids,
                        obj=qt_wallet.history_list,
                        obj_key=txid,
                    )
                    # connect also the higher parents, so the results appear at all
                    wallet_tx_ids.set_parent(wallet_item)
                    wallet_item.set_parent(root)

            wallet_utxos = ResultItem(html_f(self.tr("UTXOs"), bf=True), obj=qt_wallet.uitx_creator)
            wallet_txos = ResultItem(html_f(self.tr("Spent Outputs"), bf=True), obj=qt_wallet.history_tab)
            for pythonutxo in qt_wallet.wallet.get_all_txos_dict().values():
                outpoint_str = str(pythonutxo.outpoint)
                if search_text in outpoint_str:
                    if pythonutxo.is_spent_by_txid:
                        ResultItem(
                            format_result_text(outpoint_str),
                            parent=wallet_txos,
                            obj=qt_wallet.history_list,
                            obj_key=pythonutxo.is_spent_by_txid,
                        )
                        # connect also the higher parents, so the results appear at all
                        wallet_txos.set_parent(wallet_item)
                        wallet_item.set_parent(root)
                    else:
                        ResultItem(
                            format_result_text(outpoint_str),
                            parent=wallet_utxos,
                            obj=qt_wallet.uitx_creator.utxo_list,
                            obj_key=outpoint_str,
                        )
                        # connect also the higher parents, so the results appear at all
                        wallet_utxos.set_parent(wallet_item)
                        wallet_item.set_parent(root)

        return root


if __name__ == "__main__":

    class MainWindow(QMainWindow):
        def __init__(self) -> None:
            super().__init__()

            self.central_widget = QWidget()
            self.setCentralWidget(self.central_widget)

            self.central_widget_layout = QVBoxLayout(self.central_widget)

            self.search_tree_view = SearchTreeView(demo_do_search, on_click=demo_on_click)
            self.central_widget_layout.addWidget(self.search_tree_view)
            self.central_widget_layout.addWidget(QPushButton("dummy"))

    app = QApplication(sys.argv)
    main_window = MainWindow()
    main_window.show()
    sys.exit(app.exec())



#
# Bitcoin Safe
# Copyright (C) 2024 Andreas Griffin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/gpl-3.0.html
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import logging

logger = logging.getLogger(__name__)

import sys
from datetime import datetime, timezone

from PyQt6.QtCore import QDateTime, Qt
from PyQt6.QtWidgets import (
    QApplication,
    QCheckBox,
    QDateTimeEdit,
    QGroupBox,
    QLabel,
    QVBoxLayout,
    QWidget,
)


class DateTimePicker(QWidget):
    def __init__(self) -> None:
        super().__init__()
        self.initUI()

    def initUI(self) -> None:
        self.dateTimeEdit = QDateTimeEdit(self)
        self.dateTimeEdit.setCalendarPopup(True)

        # Set the QDateTimeEdit to the current date and time
        self.dateTimeEdit.setDateTime(QDateTime.currentDateTime())

        layout = QVBoxLayout()
        layout.addWidget(self.dateTimeEdit)
        self.setLayout(layout)

    def print_time(self) -> None:
        # Convert QDateTime to Python datetime object
        local_datetime = self.get_datetime()

        # Assuming the local_datetime is naive (no timezone information),
        # Convert it to UTC
        utc_datetime = local_datetime.astimezone(timezone.utc)
        print("UTC Time:", utc_datetime)

    def get_datetime(self) -> datetime:
        return self.dateTimeEdit.dateTime().toPyDateTime()


class CheckBoxGroupBox(QWidget):
    def __init__(self, enabled=True) -> None:
        super().__init__()
        # Create the checkbox
        self.checkbox = QCheckBox()
        self.checkbox.stateChanged.connect(self.toggleGroupBox)
        self.checkbox.setStyleSheet("margin-bottom: 0px;")  # Adjust margin as needed

        # Create the group box
        self.groupBox = QGroupBox()
        self.groupBox_layout = QVBoxLayout(self.groupBox)

        # Arrange the checkbox and group box in a layout
        layout = QVBoxLayout()
        layout.addWidget(self.checkbox)
        layout.addWidget(self.groupBox)
        self.setLayout(layout)
        self.toggleGroupBox(self.checkbox.checkState())

    def toggleGroupBox(self, state: Qt.CheckState) -> None:
        # Enable or disable the group box based on the checkbox state
        self.groupBox.setEnabled(state == Qt.CheckState.Checked)


class nLocktimePicker(CheckBoxGroupBox):
    def __init__(self) -> None:
        super().__init__()

        self.checkbox.setText("Set nLockTime")

        label = QLabel(
            'Set the minimum time (<a href="https://learn.saylor.org/mod/book/view.php?id=36369&chapterid=19000">Median-Time-Past</a>) the transaction can be included in a block. See: <a href="https://learn.saylor.org/mod/book/view.php?id=36369&chapterid=18994">nLocktime</a>'
        )
        label.setTextFormat(Qt.TextFormat.RichText)
        label.setOpenExternalLinks(True)  # Enable opening links
        label.setWordWrap(True)
        self.groupBox_layout.addWidget(label)

        self.nlocktime_picker = DateTimePicker()
        self.groupBox_layout.addWidget(self.nlocktime_picker)


if __name__ == "__main__":
    app = QApplication(sys.argv)
    window = nLocktimePicker()
    window.show()
    sys.exit(app.exec())



#
# Bitcoin Safe
# Copyright (C) 2024 Andreas Griffin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/gpl-3.0.html
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


# Original Version from:
#
# Electrum - lightweight Bitcoin client
# Copyright (C) 2015 Thomas Voegtlin
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging

from bitcoin_safe.gui.qt.wrappers import Menu

from ...config import UserConfig
from ...pythonbdk_types import OutPoint, PythonUtxo, TxOut

logger = logging.getLogger(__name__)


import enum
from typing import Dict, List, Optional, Tuple, Union

import bdkpython as bdk
from PyQt6.QtCore import QModelIndex, QPoint, Qt
from PyQt6.QtGui import QStandardItem
from PyQt6.QtWidgets import QAbstractItemView, QHeaderView, QWidget

from ...i18n import translate
from ...signals import Signals, UpdateFilter, UpdateFilterReason
from ...util import Satoshis, block_explorer_URL, clean_list
from ...wallet import TxStatus, Wallet, get_wallets
from .category_list import CategoryEditor
from .my_treeview import (
    MyItemDataRole,
    MySortModel,
    MyStandardItemModel,
    MyTreeView,
    QItemSelectionModel,
    TreeViewWithToolbar,
    needs_frequent_flag,
)
from .util import ColorScheme, read_QIcon, sort_id_to_icon, webopen


def icon_of_utxo(is_spent_by_txid: Optional[str], confirmation_time: bdk.BlockTime, sort_id: int) -> str:
    if not confirmation_time and is_spent_by_txid:
        return "unconfirmed_child.svg"
    return sort_id_to_icon(sort_id)


def tooltip_text_of_utxo(is_spent_by_txid: Optional[str], confirmation_time: bdk.BlockTime) -> str:
    if not confirmation_time:
        if is_spent_by_txid:
            return translate(
                "utxo_list", "Unconfirmed UTXO is spent by transaction {is_spent_by_txid}"
            ).format(is_spent_by_txid=is_spent_by_txid)
        else:
            return translate("utxo_list", "Unconfirmed UTXO")

    return translate("utxo_list", f"Confirmed UTXO")


class UTXOList(MyTreeView):
    class Columns(MyTreeView.BaseColumnsEnum):
        STATUS = enum.auto()
        WALLET_ID = enum.auto()
        OUTPOINT = enum.auto()
        ADDRESS = enum.auto()
        CATEGORY = enum.auto()
        LABEL = enum.auto()
        AMOUNT = enum.auto()
        PARENTS = enum.auto()

    filter_columns = [
        Columns.WALLET_ID,
        Columns.OUTPOINT,
        Columns.ADDRESS,
        Columns.CATEGORY,
        Columns.LABEL,
        Columns.AMOUNT,
    ]
    column_alignments = {
        Columns.STATUS: Qt.AlignmentFlag.AlignHCenter | Qt.AlignmentFlag.AlignVCenter,
        Columns.WALLET_ID: Qt.AlignmentFlag.AlignHCenter | Qt.AlignmentFlag.AlignVCenter,
        Columns.OUTPOINT: Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignVCenter,
        Columns.ADDRESS: Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignVCenter,
        Columns.CATEGORY: Qt.AlignmentFlag.AlignCenter | Qt.AlignmentFlag.AlignVCenter,
        Columns.LABEL: Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignVCenter,
        Columns.AMOUNT: Qt.AlignmentFlag.AlignRight | Qt.AlignmentFlag.AlignVCenter,
        Columns.PARENTS: Qt.AlignmentFlag.AlignCenter | Qt.AlignmentFlag.AlignVCenter,
    }

    column_widths: Dict[MyTreeView.BaseColumnsEnum, int] = {
        Columns.STATUS: 15,
        Columns.ADDRESS: 100,
        Columns.AMOUNT: 100,
    }
    stretch_column = Columns.LABEL
    key_column = Columns.OUTPOINT

    def __init__(
        self,
        config: UserConfig,
        signals: Signals,
        get_outpoints,
        hidden_columns: List[int] | None = None,
        txout_dict: Union[Dict[str, bdk.TxOut], Dict[str, TxOut]] | None = None,
        keep_outpoint_order=False,
        sort_column: int | None = None,
        sort_order: Qt.SortOrder | None = None,
    ):
        """_summary_

        Args:
            config (UserConfig): _description_
            signals (Signals): _description_
            get_outpoints (_type_): _description_
            hidden_columns (_type_, optional): _description_. Defaults to None.
            txout_dict (Dict[str, bdk.TxOut], optional): Can be used to augment the list with infos, if the utxo is not from the own wallet. Defaults to None.
            keep_outpoint_order (bool, optional): _description_. Defaults to False.
        """
        super().__init__(
            config=config,
            stretch_column=self.stretch_column,
            column_widths=self.column_widths,
            editable_columns=[],
            signals=signals,
            sort_column=sort_column if sort_column is not None else UTXOList.Columns.ADDRESS,
            sort_order=sort_order if sort_order is not None else Qt.SortOrder.AscendingOrder,
        )
        self.config: UserConfig = config
        self.keep_outpoint_order = keep_outpoint_order
        self.hidden_columns = hidden_columns if hidden_columns else []
        self.signals = signals
        self.get_outpoints = get_outpoints
        self.txout_dict: Union[Dict[str, bdk.TxOut], Dict[str, TxOut]] = txout_dict if txout_dict else {}
        self._pythonutxo_dict: Dict[str, PythonUtxo] = {}  # outpoint --> txdetails
        self._wallet_dict: Dict[str, Wallet] = {}  # outpoint --> wallet

        self.setTextElideMode(Qt.TextElideMode.ElideMiddle)
        self._source_model = MyStandardItemModel(self, drag_key="outpoints")
        self.proxy = MySortModel(
            self, source_model=self._source_model, sort_role=MyItemDataRole.ROLE_SORT_ORDER
        )
        self.setModel(self.proxy)

        self.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)
        self.setSortingEnabled(True)  # Allow user to sort by clicking column headers

        self.update_content()

        # signals
        signals.any_wallet_updated.connect(self.update_with_filter)
        self.signals.language_switch.connect(self.updateUi)

        # self.setDragEnabled(True)
        # self.setAcceptDrops(True)
        # self.viewport().setAcceptDrops(True)
        # self.setDropIndicatorShown(True)
        # self.setDragDropMode(QAbstractItemView.InternalMove)
        # self.setDefaultDropAction(Qt.MoveAction)

    def create_menu(self, position: QPoint) -> Menu:
        selected: List[QModelIndex] = self.selected_in_column(self.Columns.OUTPOINT)
        if not selected:
            current_row = self.current_row_in_column(self.Columns.OUTPOINT)
            if current_row:
                selected = [current_row]

        menu = Menu()

        multi_select = len(selected) > 1
        outpoints: List[OutPoint] = [
            self.model().data(item, role=MyItemDataRole.ROLE_KEY) for item in selected
        ]

        if not multi_select:
            idx = self.indexAt(position)
            if not idx.isValid():
                return menu
            item = self.item_from_index(idx)
            if not item:
                return menu

            if str(outpoints[0]) in self._wallet_dict:
                menu.add_action(
                    translate("utxo_list", "Open transaction"),
                    lambda: self.signals.open_tx_like.emit(outpoints[0].txid),
                )

            txid_URL = block_explorer_URL(self.config.network_config.mempool_url, "tx", outpoints[0].txid)
            if txid_URL:
                menu.add_action(
                    translate("utxo_list", "View on block explorer"),
                    lambda: webopen(txid_URL),
                    icon=read_QIcon("link.svg"),
                )

            wallet_ids: List[str] = clean_list(
                [
                    self.model().data(item, role=MyItemDataRole.ROLE_CLIPBOARD_DATA)
                    for item in self.selected_in_column(self.Columns.WALLET_ID)
                ]
            )
            addresses: List[str] = clean_list(
                [
                    self.model().data(item, role=MyItemDataRole.ROLE_CLIPBOARD_DATA)
                    for item in self.selected_in_column(self.Columns.ADDRESS)
                ]
            )
            if wallet_ids and addresses:
                menu.add_action(
                    translate("utxo_list", "Open Address Details"),
                    lambda: self.signals.wallet_signals[wallet_ids[0]].show_address.emit(
                        addresses[0], wallet_ids[0]
                    ),
                )

            self.add_copy_menu(menu, idx, include_columns_even_if_hidden=[self.Columns.OUTPOINT])

        menu.add_action(
            translate("utxo_list", "Copy as csv"),
            lambda: self.copyRowsToClipboardAsCSV([r.row() for r in selected]),
            icon=read_QIcon("csv-file.svg"),
        )

        # run_hook('receive_menu', menu, addrs, self.wallet)
        if viewport := self.viewport():
            menu.exec(viewport.mapToGlobal(position))

        return menu

    def get_wallet_address_satoshis(
        self, outpoint: OutPoint
    ) -> Tuple[Optional[Wallet], Optional[PythonUtxo], Optional[str], Optional[Satoshis]]:
        wallet = self._wallet_dict.get(str(outpoint))
        python_utxo = self._pythonutxo_dict.get(str(outpoint))
        address = None
        satoshis = None
        if python_utxo:
            satoshis = Satoshis(python_utxo.txout.value, self.config.network)
            address = python_utxo.address
        else:
            txout = self.txout_dict.get(str(outpoint))
            if txout:
                satoshis = Satoshis(txout.value, self.config.network)
                address = bdk.Address.from_script(txout.script_pubkey, self.config.network).as_string()
        return wallet, python_utxo, address, satoshis

    def get_headers(self):
        return {
            self.Columns.STATUS: (""),
            self.Columns.WALLET_ID: self.tr("Wallet"),
            self.Columns.OUTPOINT: self.tr("Outpoint"),
            self.Columns.ADDRESS: self.tr("Address"),
            self.Columns.CATEGORY: self.tr("Category"),
            self.Columns.LABEL: self.tr("Label"),
            self.Columns.AMOUNT: self.tr("Amount"),
            self.Columns.PARENTS: self.tr("Parents"),
        }

    def update_with_filter(self, update_filter: UpdateFilter) -> None:
        should_update = False
        if should_update or update_filter.refresh_all:
            should_update = True
        if should_update or update_filter.categories or update_filter.addresses:
            should_update = True

        if should_update:
            return self.update_content()

        logger.debug(f"{self.__class__.__name__} update_with_filter {update_filter}")

        self._before_update_content()

        log_info = []
        model = self._source_model
        # Select rows with an ID in id_list
        for row in range(model.rowCount()):
            outpoint: OutPoint = model.data(model.index(row, self.key_column))

            if (
                update_filter.reason == UpdateFilterReason.ChainHeightAdvanced
                and model.data(
                    model.index(row, self.key_column), role=MyItemDataRole.ROLE_FREQUENT_UPDATEFLAG
                )
            ) or any([outpoint in update_filter.outpoints]):
                log_info.append((row, str(outpoint)))
                self.refresh_row(outpoint, row)

        logger.debug(f"Updated  {log_info}")

        self._after_update_content()

    def update_content(self):
        if self.maybe_defer_update():
            return

        def str_format(v):
            return str(v) if v else "Unknown"

        self._before_update_content()

        current_key = self.get_role_data_for_current_item(col=self.key_column, role=MyItemDataRole.ROLE_KEY)

        # build dicts to look up the outpoints later (fast)

        self._wallet_dict = {}  # outpoint_str:Wallet
        self._pythonutxo_dict = {}  # outpoint_str:PythonUTXO
        for wallet_ in get_wallets(self.signals):
            txos_dict = wallet_.get_all_txos_dict(include_not_mine=True)
            self._pythonutxo_dict.update(txos_dict)
            self._wallet_dict.update({outpoint_str: wallet_ for outpoint_str in txos_dict.keys()})

        self._source_model.clear()
        self.update_headers(self.get_headers())
        for i, outpoint in enumerate(self.get_outpoints()):
            outpoint = OutPoint.from_bdk(outpoint)
            wallet, python_utxo, address, satoshis = self.get_wallet_address_satoshis(outpoint)

            labels = [""] * len(self.Columns)
            labels[self.Columns.OUTPOINT] = str(outpoint)
            labels[self.Columns.ADDRESS] = str_format(address)
            labels[self.Columns.AMOUNT] = str_format(satoshis)
            items = [QStandardItem(x) for x in labels]
            self.set_editability(items)
            items[self.Columns.OUTPOINT].setText(str(outpoint))
            items[self.Columns.OUTPOINT].setData(outpoint, MyItemDataRole.ROLE_KEY)
            items[self.Columns.OUTPOINT].setData(str(outpoint), MyItemDataRole.ROLE_CLIPBOARD_DATA)
            items[self.Columns.OUTPOINT].setToolTip(str(outpoint))

            items[self.Columns.ADDRESS].setData(i, MyItemDataRole.ROLE_SORT_ORDER)
            # items[self.Columns.ADDRESS].setFont(QFont(MONOSPACE_FONT))
            items[self.Columns.ADDRESS].setData(
                labels[self.Columns.ADDRESS], MyItemDataRole.ROLE_CLIPBOARD_DATA
            )
            items[self.Columns.ADDRESS].setToolTip(labels[self.Columns.ADDRESS])
            # items[self.Columns.AMOUNT].setFont(QFont(MONOSPACE_FONT))
            items[self.Columns.AMOUNT].setData(
                satoshis.value if satoshis else str_format(satoshis), MyItemDataRole.ROLE_CLIPBOARD_DATA
            )

            # add item
            count = self._source_model.rowCount()
            self._source_model.insertRow(count, items)
            self.refresh_row(outpoint, count)

        if isinstance(header := self.header(), QHeaderView):
            header.setSectionResizeMode(self.Columns.ADDRESS, QHeaderView.ResizeMode.Interactive)

        self._after_update_content()
        super().update_content()

    def refresh_row(self, key: bdk.OutPoint, row: int):
        assert row is not None

        outpoint = OutPoint.from_bdk(key)
        wallet, python_utxo, address, satoshis = self.get_wallet_address_satoshis(outpoint)
        if not python_utxo:
            return

        txdetails = wallet.get_tx(outpoint.txid) if wallet else None
        status = TxStatus.from_wallet(txdetails.txid, wallet) if txdetails and wallet else None
        sort_id = status.sort_id() if status else -1

        _items = [self._source_model.item(row, col) for col in self.Columns]
        items = [entry for entry in _items if entry]

        if needs_frequent_flag(status=status):
            # unconfirmed txos might be confirmed, and need to be updated more often
            items[self.key_column].setData(True, role=MyItemDataRole.ROLE_FREQUENT_UPDATEFLAG)
        items[self.Columns.STATUS].setIcon(
            read_QIcon(
                icon_of_utxo(python_utxo.is_spent_by_txid, txdetails.confirmation_time, sort_id)
                if txdetails
                else None
            )
        )
        if txdetails:
            items[self.Columns.STATUS].setToolTip(
                tooltip_text_of_utxo(python_utxo.is_spent_by_txid, txdetails.confirmation_time)
            )

        wallet_id = wallet.id if wallet and address and wallet.is_my_address(address) else ""
        items[self.Columns.WALLET_ID].setText(wallet_id)
        items[self.Columns.WALLET_ID].setData(wallet_id, MyItemDataRole.ROLE_CLIPBOARD_DATA)
        txid = outpoint.txid

        category = wallet.labels.get_category(address) if wallet and address else ""

        items[self.Columns.CATEGORY].setText(category if category else "")
        items[self.Columns.CATEGORY].setData(category, MyItemDataRole.ROLE_CLIPBOARD_DATA)
        label = wallet.get_label_for_txid(txid) or "" if wallet else ""
        items[self.Columns.LABEL].setText(label)
        items[self.Columns.LABEL].setData(label, MyItemDataRole.ROLE_CLIPBOARD_DATA)
        color = self._default_bg_brush
        for col in items:
            col.setBackground(color)

        items[self.Columns.CATEGORY].setBackground(CategoryEditor.color(category))
        if (
            python_utxo
            and python_utxo.txout
            and wallet
            and address
            and wallet.is_my_address(python_utxo.address)
        ):
            color = (
                ColorScheme.YELLOW.as_color(background=True)
                if wallet.is_change(address)
                else ColorScheme.GREEN.as_color(background=True)
            )
            items[self.Columns.ADDRESS].setBackground(color)

    def get_selected_outpoints(self) -> List[OutPoint]:
        if not self.model():
            return []
        items = self.selected_in_column(self.Columns.OUTPOINT)
        return [x.data(MyItemDataRole.ROLE_KEY) for x in items]

    def get_selected_values(self) -> List[int]:
        if not self.model():
            return []
        items = self.selected_in_column(self.Columns.AMOUNT)
        return [x.data(MyItemDataRole.ROLE_CLIPBOARD_DATA) for x in items]

    def on_double_click(self, idx: QModelIndex):
        outpoint = idx.sibling(idx.row(), self.Columns.OUTPOINT).data(MyItemDataRole.ROLE_KEY)
        wallets = get_wallets(self.signals)
        for wallet in wallets:
            python_utxo = wallet.get_python_txo(str(outpoint))
            if python_utxo:
                self.signals.wallet_signals[wallet.id].show_utxo.emit(outpoint)


class UtxoListWithToolbar(TreeViewWithToolbar):
    def __init__(self, utxo_list: UTXOList, config: UserConfig, parent: QWidget | None = None) -> None:
        super().__init__(utxo_list, config, parent=parent)
        self.utxo_list = utxo_list
        selection_model = self.utxo_list.selectionModel()
        if not selection_model:
            selection_model = QItemSelectionModel(self.utxo_list.model())
            self.utxo_list.setSelectionModel(selection_model)
        self.utxo_list.signal_selection_changed.connect(self.update_labels)
        self.create_layout()
        self.utxo_list.signals.language_switch.connect(self.updateUi)
        self.utxo_list.signals.any_wallet_updated.connect(self.updateUi)

    def updateUi(self):
        super().updateUi()

        self.update_labels()

    def update_labels(self):
        try:
            selected_values = self.utxo_list.get_selected_values()
            amount = sum(selected_values)
            self.uxto_selected_label.setText(
                self.tr("{amount} selected ({number} UTXOs)").format(
                    amount=Satoshis(amount, self.utxo_list.signals.get_network()).str_with_unit(),
                    number=len(selected_values),
                )
            )
        except:
            self.uxto_selected_label.setText(f"")

    def create_toolbar_with_menu(self, title):
        super().create_toolbar_with_menu(title=title)
        self.uxto_selected_label = self.balance_label



#
# Bitcoin Safe
# Copyright (C) 2024 Andreas Griffin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/gpl-3.0.html
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import logging

from bitcoin_safe.gui.qt.notification_bar import NotificationBar
from bitcoin_safe.gui.qt.util import icon_path
from bitcoin_safe.signals import SignalsMin

logger = logging.getLogger(__name__)


import bdkpython as bdk
from PyQt6.QtGui import QIcon


class NotificationBarRegtest(NotificationBar):
    def __init__(self, open_network_settings, network: bdk.Network, signals_min: SignalsMin) -> None:
        super().__init__(
            text="",
            optional_button_text="",
            callback_optional_button=open_network_settings,
            has_close_button=True,
        )
        self.network = network
        self.signals_min = signals_min
        self.set_background_color("lightblue")
        self.set_icon(QIcon(icon_path(f"bitcoin-{network.name.lower()}.svg")))

        self.updateUi()
        self.signals_min.language_switch.connect(self.updateUi)

    def updateUi(self) -> None:
        self.optionalButton.setText(self.tr("Change Network"))
        self.textLabel.setText(
            self.tr("Network = {network}. The coins are worthless!").format(
                network=self.network.name.capitalize()
            )
        )
